"""Module for communication with Knauer devices."""
import asyncio

from loguru import logger

from flowchem.utils.exceptions import InvalidConfigurationError

from .knauer_finder import autodiscover_knauer


class KnauerEthernetDevice:
    """Common base class for shared logic across Knauer pumps and valves."""

    TCP_PORT = 10001
    BUFFER_SIZE = 1024
    _id_counter = 0

    def __init__(self, ip_address, mac_address, network="", **kwargs):
        """Knauer Ethernet Device - either pump or valve.

        If a MAC address is given, it is used to autodiscover the IP address.
        Otherwise, the IP address must be given.

        Note that for configuration files, the MAC address is preferred as it is static.

        Args:
        ----
            ip_address: device IP address (only 1 of either IP or MAC address is needed)
            mac_address: device MAC address (only 1 of either IP or MAC address is needed)
            name: name of device (optional)
        """
        super().__init__(**kwargs)

        # MAC address
        if mac_address:
            self.ip_address = self._ip_from_mac(mac_address.lower(), network=network)
        else:
            self.ip_address = ip_address

        # These will be set in initialize()
        self._reader: asyncio.StreamReader = None  # type: ignore
        self._writer: asyncio.StreamWriter = None  # type: ignore

        # Note: the pump requires "\n\r" as EOL, the valves "\r\n"! So this is set by the subclasses
        self.eol = b""

        # Lock communication between write and read reply
        self._lock = asyncio.Lock()

    def _ip_from_mac(self, mac_address: str, network="") -> str:
        """Get IP from MAC."""
        # Autodiscover IP from MAC address
        available_devices = autodiscover_knauer(network)
        # IP if found, None otherwise
        ip_address = available_devices.get(mac_address)
        if ip_address is None:
            raise InvalidConfigurationError(
                f"{self.__class__.__name__}:{self.name}\n"  # type: ignore
                f"Device with MAC address={mac_address} not found!\n"
                f"[Available: {available_devices}]"
            )
        return ip_address

    async def initialize(self):
        """Initialize connection."""
        # Future used to set shorter timeout than default
        future = asyncio.open_connection(host=self.ip_address, port=10001)
        try:
            self._reader, self._writer = await asyncio.wait_for(future, timeout=3)
        except OSError as connection_error:
            logger.exception(connection_error)
            raise InvalidConfigurationError(
                f"Cannot open connection with device {self.__class__.__name__} at IP={self.ip_address}"
            ) from connection_error
        except asyncio.TimeoutError as timeout_error:
            logger.exception(timeout_error)
            raise InvalidConfigurationError(
                f"No reply from device {self.__class__.__name__} at IP={self.ip_address}"
            ) from timeout_error

    async def _send_and_receive(self, message: str) -> str:
        async with self._lock:
            self._writer.write(message.encode("ascii") + self.eol)
            await self._writer.drain()
            logger.debug(f"WRITE >>> '{message}' ")
            reply = await self._reader.readuntil(separator=b"\r")
        logger.debug(f"READ <<< '{reply.decode().strip()}' ")
        return reply.decode("ascii").strip()



"""Control module for the Custom Peltier cooler components."""
from __future__ import annotations

from typing import TYPE_CHECKING
from flowchem.components.technical.temperature import TemperatureControl, TempRange


if TYPE_CHECKING:
    from .peltier_cooler import PeltierCooler

class PeltierCoolerTemperatureControl(TemperatureControl):
    """Peltier Cooler ."""

    hw_device: PeltierCooler  # for typing's sake

    def __init__(
        self, name: str, hw_device: PeltierCooler, temp_limits: TempRange
    ) -> None:
        """Create a TemperatureControl object."""
        super().__init__(name, hw_device, temp_limits)

    async def set_temperature(self, temperature: str):
        """Set the target temperature to the given string in "magnitude and unit" format."""
        set_t = await super().set_temperature(temperature)
        return await self.hw_device.set_temperature(set_t)

    async def get_temperature(self) -> float:  # type: ignore
        """Return temperature in Celsius."""
        return await self.hw_device.get_temperature()

    async def is_target_reached(self) -> bool | None:  # type: ignore
        """Return True if the set temperature target has been reached."""
        current_temp = await self.hw_device.get_temperature()
        params = await self.hw_device.get_parameters()
        values = params.split(',')
        target_temp = float(values[0])
        if abs(current_temp - target_temp) <= 2:
            return True
        else:
            return False

    async def power_on(self):
        """Turn on temperature control."""
        return await self.hw_device.start_control()

    async def power_off(self):
        """Turn off temperature control."""
        return await self.hw_device.stop_control()

    async def temperature_limits(self) -> TempRange:
        """Return a dict with `min` and `max` temperature in Celsius."""
        return self._limits



"""An IR control component."""
from pydantic import BaseModel

from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.devices.flowchem_device import FlowchemDevice


class IRSpectrum(BaseModel):
    """IR spectrum class.

    Consider rampy for advance features (baseline fit, etc.)
    See e.g. https://github.com/charlesll/rampy/blob/master/examples/baseline_fit.ipynb
    """

    wavenumber: list[float]
    intensity: list[float]


class IRControl(FlowchemComponent):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """HPLC Control component. Sends methods, starts run, do stuff."""
        super().__init__(name, hw_device)
        self.add_api_route("/acquire-spectrum", self.acquire_spectrum, methods=["PUT"])
        self.add_api_route("/stop", self.stop, methods=["PUT"])

        # Ontology: high performance liquid chromatography instrument
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/OBI_0001057",
        )
        self.component_info.type = "IR Control"

    async def acquire_spectrum(self) -> IRSpectrum:  # type: ignore
        """Acquire an IR spectrum."""
        ...

    async def stop(self):
        """Stop acquisition and exit gracefully."""
        ...



from __future__ import annotations
import asyncio
import enum
import threading
from io import BytesIO
from pathlib import Path
from typing import Any

from loguru import logger

from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.server.configuration_parser import (
    instantiate_device_from_config,
    parse_config,
)
from flowchem.server.fastapi_server import FastAPIServer
from flowchem.server.zeroconf_server import ZeroconfServer


class _Flowchem(threading.local):
    """Container which makes a Flowchem instance available to the event loop."""

    fc: Flowchem | None = None


# Essentially a global pointer to flowchem as thread local
_fc = _Flowchem()


class CoreState(enum.Enum):
    """Represent the current state of Flowchem."""

    not_running = "NOT_RUNNING"
    starting = "STARTING"
    running = "RUNNING"
    stopping = "STOPPING"
    final_write = "FINAL_WRITE"
    stopped = "STOPPED"

    def __str__(self) -> str:
        """Return the event."""
        return self.value


class Flowchem:
    def __new__(cls) -> Flowchem:
        """Set the _fc thread local data."""
        fc = super().__new__(cls)
        _fc.fc = fc
        return fc

    def __init__(self):
        self.loop = asyncio.get_running_loop()
        self._tasks: set[asyncio.Future[Any]] = set()
        self.config: dict[str, Any] = {}
        self.devices: list[FlowchemDevice] = []
        self.state: CoreState = CoreState.not_running
        self.exit_code: int = 0
        # If not None, use to signal end-of-loop
        self._stopped: asyncio.Event | None = None

        # mDNS server (Zeroconf)
        self.mdns = ZeroconfServer(self.port)

        # HTTP server (FastAPI)
        self.http = FastAPIServer(
            self.config.get("filename", ""),
            host=self.mdns.mdns_addresses[0],
            port=self.port,
        )

        # To be implemented
        # self.bus = EventBus(self)
        # self.states = StateMachine(self.bus, self.loop)

    @property
    def port(self):
        return self.config.get("port", 8000)

    async def setup(self, config: BytesIO | Path):
        self.config = parse_config(config)
        self.devices = instantiate_device_from_config(self.config)

        """Initialize connection to devices and create API endpoints."""
        logger.info("Initializing device connection(s)...")

        # Run `initialize` async method of all hw devices in parallel
        await asyncio.gather(*[dev.initialize() for dev in self.devices])
        logger.info("Device(s) connected")

        # Create entities for the configured devices.
        for device in self.devices:
            # Advertise devices as services via mDNS
            await self.mdns.add_device(name=device.name)
            # Add device API to HTTP server
            self.http.add_device(device)
        logger.info("Server component(s) loaded successfully!")


if __name__ == "__main__":
    import uvicorn

    async def main():
        flowchem = Flowchem()
        await flowchem.setup(
            BytesIO(b"""[device.test-device]\ntype = "FakeDevice"\n""")
        )

        config = uvicorn.Config(
            flowchem.http.app,
            port=flowchem.port,
            log_level="info",
            timeout_keep_alive=3600,
        )
        server = uvicorn.Server(config)
        await server.serve()

    asyncio.run(main())



"""Power control, sets both voltage and current. (Could be split in two, unnecessary for now)."""
from __future__ import annotations

from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.devices.flowchem_device import FlowchemDevice


class PowerSwitch(FlowchemComponent):
    """A generic power on/off switch."""

    def __init__(
        self,
        name: str,
        hw_device: FlowchemDevice,
    ) -> None:
        """Create a TemperatureControl object."""
        super().__init__(name, hw_device)

        self.add_api_route("/power-on", self.power_on, methods=["PUT"])
        self.add_api_route("/power-off", self.power_off, methods=["PUT"])

    async def power_on(self):
        """Turn power on."""
        ...

    async def power_off(self):
        """Turn off power."""
        ...


class PowerControl(PowerSwitch):
    """A generic power controller, adjusting voltage and/or current."""

    def __init__(
        self,
        name: str,
        hw_device: FlowchemDevice,
    ) -> None:
        """Create a PowerSwitch object."""
        super().__init__(name, hw_device)

        self.add_api_route("/power-on", self.power_on, methods=["PUT"])
        self.add_api_route("/power-off", self.power_off, methods=["PUT"])

        # Current
        self.add_api_route("/current", self.get_current, methods=["GET"])
        self.add_api_route("/current", self.set_current, methods=["PUT"])

        # Voltage
        self.add_api_route("/voltage", self.get_voltage, methods=["GET"])
        self.add_api_route("/voltage", self.set_voltage, methods=["PUT"])

    async def set_current(self, current: str):
        """Set the target current to the given string in "magnitude and unit" format."""
        ...

    async def get_current(self) -> float:  # type: ignore
        """Return current in Ampere."""
        ...

    async def set_voltage(self, voltage: str):
        """Set the target voltage to the given string in "magnitude and unit" format."""
        ...

    async def get_voltage(self) -> float:  # type: ignore
        """Return current in Volt."""
        ...



"""Parse a device config file."""
import inspect
import sys
import typing
from io import BytesIO
from pathlib import Path

from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.list_known_device_type import autodiscover_device_classes

if sys.version_info >= (3, 11):
    # noinspection PyUnresolvedReferences
    import tomllib
else:
    import tomli as tomllib

from loguru import logger

from flowchem.devices.known_plugins import plugin_devices
from flowchem.utils.exceptions import InvalidConfigurationError

DEVICE_NAME_MAX_LENGTH = 42


def parse_toml(stream: typing.BinaryIO) -> dict:
    """Read the TOML configuration file and returns it as a dict.

    Extensive exception handling due to the error-prone human editing needed in the configuration file.
    """
    try:
        return tomllib.load(stream)
    except tomllib.TOMLDecodeError as parser_error:
        logger.exception(parser_error)
        msg = "Invalid syntax in configuration!"
        raise InvalidConfigurationError(msg) from parser_error


def parse_config(file_path: Path | BytesIO) -> dict:
    """Parse a config file and add a `filename` key to the resulting dict w/ its location."""
    # BytesIO used for testing without creating actual files
    if isinstance(file_path, BytesIO):
        config = parse_toml(file_path)
        config["filename"] = "BytesIO"
    elif isinstance(file_path, Path):
        assert file_path.exists(), f"{file_path} exists"
        assert file_path.is_file(), f"{file_path} is a file"

        with file_path.open("rb") as stream:
            config = parse_toml(stream)
        config["filename"] = file_path.stem
    else:
        raise InvalidConfigurationError("Invalid configuration type.")

    return config


def instantiate_device_from_config(config: dict) -> list[FlowchemDevice]:
    """Instantiate all devices defined in the provided config dict."""
    assert "device" in config, "The configuration file must include a device section"

    # device_mapper is a dict mapping device type (str, as key) with the device class (obj, value).
    # e.g. device_mapper["Spinsolve"] = Spinsolve class
    device_mapper = autodiscover_device_classes()

    # Iterate on all devices, parse device-specific settings and instantiate the relevant objects
    return [
        parse_device(dev_settings, device_mapper)
        for dev_settings in config["device"].items()
    ]


def ensure_device_name_is_valid(device_name: str) -> None:
    """Device name validator.

    Uniqueness of names is ensured by their toml dict key nature,
    """
    if len(device_name) > DEVICE_NAME_MAX_LENGTH:
        # f"{name}._labthing._tcp.local." has to be shorter than 64 characters to be valid for mDNS
        msg = f"Device name '{device_name}' is too long: {len(device_name)} characters, max is {DEVICE_NAME_MAX_LENGTH}"
        raise InvalidConfigurationError(msg)
    if "." in device_name:
        # This is not strictly needed but avoids potential zeroconf problems
        msg = f"Invalid character '.' in device name '{device_name}'"
        raise InvalidConfigurationError(msg)


def parse_device(dev_settings: dict, device_object_mapper: dict) -> FlowchemDevice:
    """Parse device config and return a device object.

    The device type (i.e. domain) is searched in the known classes (and known plugins, even if uninstalled).
    The device is instantiated via its `from_config` method or `init` if from_config is missing.
    The device object is returned.

    Exception handling to provide more specific and diagnostic messages upon errors in the configuration file.
    """
    device_name, device_config = dev_settings
    ensure_device_name_is_valid(device_name)

    # Get device class
    try:
        obj_type = device_object_mapper[device_config["type"]]
        del device_config["type"]
    except KeyError as error:
        # If the device type specified is supported via a plugin we know of, alert user
        if device_config["type"] in plugin_devices:
            needed_plugin = plugin_devices[device_config["type"]]
            logger.exception(
                f"The device `{device_name}` of type `{device_config['type']}` needs a additional plugin"
                f"Install {needed_plugin} to add support for it!"
                f"e.g. `python -m pip install {needed_plugin}`",
            )
            msg = f"{needed_plugin} not installed."
            raise InvalidConfigurationError(msg) from error

        logger.exception(
            f"Device type `{device_config['type']}` unknown in 'device.{device_name}'!"
            f"[Known types: {device_object_mapper.keys()}]",
        )
        msg = f"Unknown device type `{device_config['type']}`."
        raise InvalidConfigurationError(msg) from error

    # If the object has a 'from_config' method, use that for instantiation, otherwise try straight with the constructor.
    try:
        if hasattr(obj_type, "from_config"):
            called = obj_type.from_config
            device = obj_type.from_config(**device_config, name=device_name)
        else:
            called = obj_type.__init__
            device = obj_type(**device_config, name=device_name)
    except TypeError as error:
        logger.error(f"Wrong settings for device '{device_name}'!")
        get_helpful_error_message(device_config, inspect.getfullargspec(called))
        msg = (
            f"Wrong configuration provided for device '{device_name}' of type {obj_type.__name__}!\n"
            f"Configuration: {device_config}\n"
            f"Accepted parameters: {inspect.getfullargspec(called).args}"
        )

        raise ConnectionError(msg) from error

    logger.debug(f"Created '{device.name}' instance: {device.__class__.__name__}")
    return device


def get_helpful_error_message(called_with: dict, arg_spec: inspect.FullArgSpec) -> None:
    """Give helpful debugging text on configuration errors."""
    # First check if we have provided an argument that is not supported.
    # Clearly no **kwargs should be defined in the signature otherwise all kwargs are ok
    if arg_spec.varkw is None:
        invalid_parameters = list(set(called_with.keys()).difference(arg_spec.args))
        if invalid_parameters:
            logger.error(
                f"The following parameters were not recognized: {invalid_parameters}",
            )

    # Then check if a mandatory arguments was not satisfied. [1 to skip cls/self, -n to remove args w/ default]
    num_default = 0 if arg_spec.defaults is None else len(arg_spec.defaults)
    mandatory_args = arg_spec.args[1:-num_default]
    missing_parameters = list(set(mandatory_args).difference(called_with.keys()))
    if missing_parameters:
        logger.error(
            f"The following mandatory parameters were missing in the configuration: {missing_parameters}",
        )



"""Spinsolve module."""
import asyncio
import pprint as pp
import warnings
from pathlib import Path

from fastapi import BackgroundTasks
from loguru import logger
from lxml import etree
from packaging import version

from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.magritek._msg_maker import (
    create_message,
    create_protocol_message,
    get_request,
    set_attribute,
    set_data_folder,
)
from flowchem.devices.magritek._parser import (
    StatusNotification,
    parse_status_notification,
)
from flowchem.devices.magritek.spinsolve_control import SpinsolveControl
from flowchem.devices.magritek.utils import create_folder_mapper, get_my_docs_path
from flowchem.utils.people import dario, jakob, wei_hsin

__all__ = ["Spinsolve"]


class Spinsolve(FlowchemDevice):
    """Spinsolve class, gives access to the spectrometer remote control API"""

    def __init__(
        self,
        host="127.0.0.1",
        port: int | None = 13000,
        name: str | None = None,
        xml_schema=None,
        data_folder=None,
        solvent: str | None = "Chloroform-d1",
        sample_name: str | None = "Unnamed automated experiment",
        remote_to_local_mapping: list[str] | None = None,
    ) -> None:
        """Control a Spinsolve instance via HTTP XML API."""
        super().__init__(name)

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Magritek",
            model="Spinsolve",
        )

        self.host, self.port = host, port

        # The Qs must exist before the response is received to be awaited
        # Could be generated programmatically from RemoteControl.xsd, but it's not worth it
        self._replies_by_type: dict[str, asyncio.Queue] = {
            "HardwareResponse": asyncio.Queue(),
            "AvailableProtocolOptionsResponse": asyncio.Queue(),
            "GetResponse": asyncio.Queue(),
            "StatusNotification": asyncio.Queue(),
        }

        # Set experimental variable
        self._data_folder = data_folder

        # An optional mapping between remote and local folder location can be used for remote use
        if remote_to_local_mapping is not None:
            self._folder_mapper = create_folder_mapper(*remote_to_local_mapping)
            assert (
                self._folder_mapper(self._data_folder) is not None
            )  # Ensure mapper validity.
        else:
            self._folder_mapper = None  # type: ignore

        # Sets default sample, solvent value and user data
        self.sample, self.solvent = sample_name, solvent
        self.protocols: dict[str, dict] = {}
        self.user_data = {"control_software": "flowchem"}

        # XML schema for reply validation. Reply validation is completely optional!
        if xml_schema is None:
            # This is the default location upon Spinsolve installation.
            # However, remote control can be from remote, i.e. not Spinsolve PC so this ;)
            default_schema = (
                get_my_docs_path() / "Magritek" / "Spinsolve" / "RemoteControl.xsd"
            )
            try:
                self.schema = etree.XMLSchema(file=str(default_schema))
            except etree.XMLSchemaParseError as et:  # i.e. not found
                raise ConnectionError(
                    f"Cannot find RemoteControl.xsd in {default_schema}!"
                ) from et
        else:
            self.schema = xml_schema

        # IOs (these are set upon initialization w/ initialize)
        self._io_reader: asyncio.StreamReader = None  # type: ignore
        self._io_writer: asyncio.StreamWriter = None  # type: ignore
        self.reader: asyncio.Task = None  # type: ignore
        # Each protocol adds a new Path to the list, run_protocol returns the ID of the next protocol
        self._result_folders: list[Path] = []

    async def initialize(self):
        """Initiate connection with a running Spinsolve instance."""
        try:
            self._io_reader, self._io_writer = await asyncio.open_connection(
                self.host,
                self.port,
            )
            logger.debug(f"Connected to {self.host}:{self.port}")
        except OSError as e:
            raise ConnectionError(
                f"Error connecting to {self.host}:{self.port} -- {e}"
            ) from e

        # Start reader thread
        self.reader = asyncio.create_task(
            self.connection_listener(),
            name="Connection listener",
        )

        # This request is used to check if the instrument is connected
        hw_info = await self.hw_request()
        if hw_info.find(".//ConnectedToHardware").text != "true":
            raise ConnectionError("Spectrometer not connected to Spinsolve PC!")

        # If connected parse and log instrument info
        self.device_info.version = hw_info.find(".//SpinsolveSoftware").text
        hardware_type = hw_info.find(".//SpinsolveType").text
        self.device_info.additional_info["hardware_type"] = hardware_type
        logger.debug(
            f"Connected to model {hardware_type}, SW: {self.device_info.version}",
        )

        # Load available protocols
        await self.load_protocols()

        # Finally, check version
        if version.parse(self.device_info.version) < version.parse("1.18.1.3062"):
            warnings.warn(
                f"Spinsolve v. {self.device_info.version} is not supported!"
                f"Upgrade to a more recent version! (at least 1.18.1.3062)",
            )

        await self.set_data_folder(self._data_folder)

        self.components.append(SpinsolveControl("nmr-control", self))

    async def connection_listener(self):
        """Listen for replies and puts them in the queue."""
        logger.debug("Spinsolve connection listener started!")
        parser = etree.XMLParser()
        while True:
            raw_tree = await self._io_reader.readuntil(b"</Message>")
            logger.debug(f"Read reply {raw_tree!r}")

            # Try parsing reply and skip if not valid
            try:
                parsed_tree = etree.fromstring(raw_tree, parser)
            except etree.XMLSyntaxError:
                warnings.warn(f"Cannot parse response XML {raw_tree!r}")
                continue

            # Validate reply if schema was provided
            if self.schema:
                try:
                    self.schema.validate(parsed_tree)
                except etree.XMLSyntaxError as syntax_error:
                    warnings.warn(
                        f"Invalid XML received! [Validation error: {syntax_error}]",
                    )

            # Add to reply queue of the given tag-type
            await self._replies_by_type[parsed_tree[0].tag].put(parsed_tree) # type: ignore

    async def get_solvent(self) -> str:
        """Get current solvent."""
        await self.send_message(get_request("Solvent"))
        reply = await self._replies_by_type["GetResponse"].get()
        return reply.find(".//Solvent").text

    async def set_solvent(self, solvent: str):
        """Set solvent."""
        await self.send_message(set_attribute("Solvent", solvent))

    async def get_sample(self) -> str:
        """Get current sample."""
        await self.send_message(get_request("Sample"))
        reply = await self._replies_by_type["GetResponse"].get()
        return reply.find(".//Sample").text

    async def set_sample(self, sample: str):
        """Set the sample name (it will appear in `acqu.par`)."""
        await self.send_message(set_attribute("Sample", sample))

    async def set_data_folder(self, location: str):
        """Set location of the data folder (where FIDs are saved to)."""
        if location is not None:
            self._data_folder = location
            await self.send_message(set_data_folder(location))

    async def get_user_data(self) -> dict:
        """Get user data. These will appear in `acqu.par`."""
        await self.send_message(get_request("UserData"))
        reply = await self._replies_by_type["GetResponse"].get()
        return {
            data_item.get("key"): data_item.get("value")
            for data_item in reply.findall(".//Data")
        }

    async def set_user_data(self, data: dict):
        """Set user data. The items provide will appear in `acqu.par`."""
        user_data = set_attribute("UserData")
        node = user_data.find(".//UserData")
        assert isinstance(node, etree._Element)
        for key, value in data.items():
            etree.SubElement(node, "Data", {key: value})
        await self.send_message(user_data)

    async def _transmit(self, message: bytes):
        """Send the message to the spectrometer."""
        # This assertion is here for mypy ;)
        assert isinstance(
            self._io_writer,
            asyncio.StreamWriter,
        ), "The connection was not initialized!"
        self._io_writer.write(message)
        await self._io_writer.drain()

    async def send_message(self, root: etree._Element):
        """Send the tree connected XML etree.Element provided."""
        # Turn the etree.Element provided into an ElementTree
        tree = etree.ElementTree(root)

        # The request must include the XML declaration
        message = etree.tostring(tree, xml_declaration=True, encoding="utf-8")

        # Send request
        logger.debug(f"Transmitting request to spectrometer: {message!r}")
        await self._transmit(message)

    async def hw_request(self):
        """Send an HW request to the spectrometer, receive the reply and returns it."""
        await self.send_message(create_message("HardwareRequest"))
        return await self._replies_by_type["HardwareResponse"].get()

    async def load_protocols(self):
        """Get a list of available protocol on the current spectrometer."""
        await self.send_message(create_message("AvailableProtocolOptionsRequest"))
        reply = await self._replies_by_type["AvailableProtocolOptionsResponse"].get()

        # Parse reply and construct the dict with protocols available
        for element in reply.findall(".//Protocol"):
            protocol_name = element.get("protocol")
            self.protocols[protocol_name] = {
                option.get("name"): [value.text for value in option.findall("Value")]
                for option in element.findall("Option")
            }

    async def run_protocol(
        self,
        name,
        background_tasks: BackgroundTasks,
        options=None,
    ) -> int:
        """Run a protocol.

        Return the ID of the protocol (needed to get results via `get_result_folder`). -1 for errors.
        """
        # All protocol names are UPPERCASE, so force upper here to avoid case issues
        name = name.upper()
        if name not in self.protocols:
            warnings.warn(
                f"The protocol requested '{name}' is not available on the spectrometer!\n"
                f"Valid options are: {pp.pformat(sorted(self.protocols.keys()))}",
            )
            return -1

        # Validate protocol options (check values and remove invalid ones, with warning)
        options_validated = self._validate_protocol_request(name, options)

        # Flush all previous StatusNotification replies from queue.
        # This is needed as sometimes FINISHED is received before other notification that remain unconsumed
        for _ in range(self._replies_by_type["StatusNotification"].qsize()):
            self._replies_by_type["StatusNotification"].get_nowait()

        # Start protocol
        await self.send_message(create_protocol_message(name, options_validated))

        # Rest of protocol as bg task to avoid timeout on API reply
        # See https://fastapi.tiangolo.com/tutorial/background-tasks/
        background_tasks.add_task(self._check_notifications)

        return len(self._result_folders)

    async def _check_notifications(self):
        """Read all the StatusNotification and returns the dataFolder."""
        self._protocol_running = True
        remote_folder = Path()
        while True:
            # Get all StatusNotification
            status_update = await self._replies_by_type["StatusNotification"].get()

            # Parse them
            status, folder = parse_status_notification(status_update)
            logger.debug(f"Status update: Status is {status} and data folder={folder}")

            # When I get a finishing response end protocol and return the data folder!
            if status is StatusNotification.FINISHING:
                remote_folder = Path(folder)
                break

            if status is StatusNotification.ERROR:
                # Usually device busy
                warnings.warn("Error detected on running protocol -- aborting.")
                await self.abort()  # Abort running experiment
                break

        logger.info(f"Protocol over - remote data folder is {remote_folder}")
        # Add result folder to self._result_folders
        if self._folder_mapper is not None:
            self._result_folders.append(self._folder_mapper(remote_folder))
        else:
            self._result_folders.append(remote_folder)

        self._protocol_running = False

    async def is_protocol_running(self) -> bool:
        """Return True if a protocol is running, otherwise False."""
        return self._protocol_running

    async def get_result_folder(self, result_id: int | None = None) -> str:
        """Get the result folder with the given ID or the last one if no ID is specified. Empty str if not existing."""
        # If no result_id get last
        if result_id is None:
            result_id = -1
        try:
            folder = str(self._result_folders[result_id])
        except IndexError:
            folder = ""

        return folder

    async def abort(self):
        """Abort running command."""
        await self.send_message(create_message("Abort"))

    def list_protocols(self) -> list[str]:
        """Return known protocol names."""
        return list(self.protocols.keys())

    def _validate_protocol_request(self, protocol_name, protocol_options) -> dict:
        """Ensure the validity of protocol name and options based on spectrometer reported parameters."""
        # Valid option for protocol
        valid_options = self.protocols.get(protocol_name)
        if valid_options is None or protocol_options is None:
            return {}

        # For each option, check if valid. If not, remove it, raise warning and continue
        for option_name, option_value in list(protocol_options.items()):
            if option_name not in valid_options:
                protocol_options.pop(option_name)
                warnings.warn(
                    f"Invalid option {option_name} for protocol {protocol_name} -- DROPPED!",
                )
                continue

            # Get valid option values (list of them or empty list if not a multiple choice)
            valid_values = valid_options[option_name]

            # If there is no list of valid options accept anything
            if not valid_values:
                continue
            # otherwise validate the value as well
            elif str(option_value) not in valid_values:
                protocol_options.pop(option_name)
                warnings.warn(
                    f"Invalid value {option_value} for option {option_name} in protocol {protocol_name}"
                    f" -- DROPPED!",
                )

        # Returns the dict with only valid options/value pairs
        return protocol_options

    def shim(self):
        """Shim on sample."""
        raise NotImplementedError("Use run protocol with a shimming protocol instead!")


if __name__ == "__main__":

    async def main():
        """Test connection."""
        nmr: Spinsolve = Spinsolve()
        await nmr.initialize()
        s = await nmr.get_solvent()
        print(f"The current solvent is set to '{s}'.")

    asyncio.run(main())



import time

from loguru import logger
from zeroconf import ServiceBrowser, Zeroconf

from flowchem.client.common import (
    FLOWCHEM_TYPE,
    FlowchemCommonDeviceListener,
    device_url_from_service_info,
    flowchem_devices_from_url_dict,
    zeroconf_name_to_device_name,
)
from flowchem.client.device_client import FlowchemDeviceClient


class FlowchemDeviceListener(FlowchemCommonDeviceListener):
    """Listener for Zeroconf service browser."""
    def _save_device_info(self, zc: Zeroconf, type_: str, name: str, active_ips: list | None = None) -> None:
        if service_info := zc.get_service_info(type_, name):
            device_name = zeroconf_name_to_device_name(name)
            if url := device_url_from_service_info(service_info, device_name, active_ips):
                self.flowchem_devices[device_name] = url
        else:
            logger.warning(f"No info for service {name}!")


def get_all_flowchem_devices(timeout: float = 3000) -> dict[str, FlowchemDeviceClient]:
    """Search for flowchem devices and returns them in a dict (key=name, value=IPv4Address)."""
    listener = FlowchemDeviceListener()
    browser = ServiceBrowser(Zeroconf(), FLOWCHEM_TYPE, listener)
    time.sleep(timeout / 1000)
    browser.cancel()

    return flowchem_devices_from_url_dict(listener.flowchem_devices)


if __name__ == "__main__":
    flowchem_devices: dict[str, FlowchemDeviceClient] = get_all_flowchem_devices()
    print(flowchem_devices)



from pydantic import AnyHttpUrl, BaseModel, NameEmail

from flowchem import __version__


class DeviceInfo(BaseModel):
    """Metadata associated with hardware devices."""

    manufacturer: str = ""
    model: str = ""
    version: str = ""
    serial_number: str | int = "unknown"
    components: dict[str, AnyHttpUrl] = {}
    backend: str = f"flowchem v. {__version__}"
    authors: list[NameEmail] = []
    additional_info: dict = {}



"""Use Phidgets to control lab devices. So far, only 4..20mA interface for Swagelock Pressure-sensor."""
import time

import pint
from loguru import logger

from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.phidgets.pressure_sensor_component import (
    PhidgetPressureSensorComponent,
)
from flowchem.utils.people import dario, jakob, wei_hsin

try:
    from Phidget22.Devices.CurrentInput import CurrentInput, PowerSupply
    from Phidget22.PhidgetException import PhidgetException

    HAS_PHIDGET = True
except ImportError:
    HAS_PHIDGET = False


from flowchem import ureg
from flowchem.utils.exceptions import InvalidConfigurationError


class PhidgetPressureSensor(FlowchemDevice):
    """Use a Phidget current input to translate a Swagelock 4..20mA signal to the corresponding pressure value."""

    def __init__(
        self,
        pressure_range: tuple[str, str] = ("0 bar", "10 bar"),
        vint_serial_number: int = -1,
        vint_channel: int = -1,
        phidget_is_remote: bool = False,
        name: str = "",
    ) -> None:
        """Initialize PressureSensor with the given pressure range (sensor-specific!)."""
        super().__init__(name=name)
        if not HAS_PHIDGET:
            raise InvalidConfigurationError(
                "Phidget unusable: library or package not installed."
            )

        # Sensor range
        sensor_min, sensor_max = pressure_range
        self._min_pressure = ureg.Quantity(sensor_min)
        self._max_pressure = ureg.Quantity(sensor_max)
        # current meter
        self.phidget = CurrentInput()

        # Ensure connection with the right sensor (ideally these are from config)
        if vint_serial_number > -1:
            self.phidget.setDeviceSerialNumber(vint_serial_number)
        if vint_channel > -1:
            self.phidget.setChannel(vint_channel)

        # Fancy remote sensors?
        if phidget_is_remote:
            from Phidget22.Net import Net
            from Phidget22.PhidgetServerType import PhidgetServerType

            Net.enableServerDiscovery(PhidgetServerType.PHIDGETSERVER_DEVICEREMOTE)
            self.phidget.setIsRemote(True)

        try:
            self.phidget.openWaitForAttachment(1000)
            logger.debug("Pressure sensor connected!")
        except PhidgetException as pe:
            raise InvalidConfigurationError(
                "Cannot connect to sensor! Check it is not already opened elsewhere and settings..."
            ) from pe

        # Set power supply to 24V
        self.phidget.setPowerSupply(PowerSupply.POWER_SUPPLY_24V)
        self.phidget.setDataInterval(200)  # 200ms

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Phidget",
            model="VINT",
            serial_number=vint_serial_number,
        )

    async def initialize(self):

        self.components.extend([PhidgetPressureSensorComponent("pressure-sensor", self)])

    def __del__(self) -> None:
        """Ensure connection closure upon deletion."""
        self.phidget.close()

    def is_attached(self) -> bool:
        """Whether the device is connected."""
        return bool(self.phidget.getAttached())

    def _current_to_pressure(self, current_in_ampere: float) -> pint.Quantity:
        """Convert current reading into pressure value."""
        mill_amp = current_in_ampere * 1000
        # minP..maxP is 4..20mA
        pressure_reading = self._min_pressure + ((mill_amp - 4) / 16) * (
            self._max_pressure - self._min_pressure
        )
        logger.debug(f"Read pressure {pressure_reading}!")
        return pressure_reading

    def read_pressure(self) -> pint.Quantity:  # type: ignore
        """Read pressure from the sensor and returns it as pint.Quantity.

        This is the main class method, and it never fails, but rather return None. Why?

        Well, initialization exception are fair play, upon startup any misconfiguration
        should be addressed.
        However, during experiment execution temporarily unavailability of a datasource
        should not be critical when the sensor is readonly.
        If the P-sensor is safety-critical than an event handler should be attached to it.
        If not we can live with it, returning None and letting the caller decide what
        to do with that.
        """
        try:
            current = self.phidget.getCurrent()
            logger.debug(f"Actual current: {current}")
        except PhidgetException:
            logger.error("Cannot read pressure!")
            return 0 * ureg.bar
        else:
            return self._current_to_pressure(current)


if __name__ == "__main__":
    test = PhidgetPressureSensor(
        pressure_range=("0 bar", "25 bar"),
        vint_serial_number=627768,
        vint_channel=0,
    )
    while True:
        print(test.read_pressure())
        time.sleep(1)



"""Vacuubrand CVC3000 control."""
import asyncio

import aioserial
import pint
from loguru import logger

from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.vacuubrand.cvc3000_pressure_control import CVC3000PressureControl
from flowchem.devices.vacuubrand.constants import ProcessStatus
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import dario, jakob, wei_hsin


class CVC3000(FlowchemDevice):
    """
    Control class for the Vacuubrand CVC3000 vacuum controller.

    This class provides methods to interface with and control the Vacuubrand CVC3000 vacuum controller via serial
    commands. It includes functionalities for setting and getting pressure, controlling motor speed, and querying
    the device status.

    Attributes:
    -----------
    DEFAULT_CONFIG : dict
        Default configuration parameters for the serial connection.
    _serial : aioserial.AioSerial
        The serial interface used to communicate with the device.
    _device_sn : int
        The serial number of the device (initialized as None).
    device_info : DeviceInfo
        Metadata and configuration details about the device.

    Methods:
    --------
    from_config(cls, port: str, name: str = None, **serial_kwargs) -> CVC3000:
        Create an instance from configuration parameters.
    initialize(self) -> None:
        Initialize the connection with the device and configure it.
    _send_command_and_read_reply(self, command: str) -> str:
        Send a command to the device and read the reply.
    version(self) -> str:
        Retrieve the version of the CVC3000 device.
    set_pressure(self, pressure: pint.Quantity) -> None:
        Set the pressure on the device.
    get_pressure(self) -> float:
        Get the current pressure from the device.
    motor_speed(self, speed: int) -> None:
        Set the motor speed on the device (0 to 100%).
    status(self) -> ProcessStatus:
        Get the current process status from the device.
    """

    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 19200,  # Supports 2400-19200
        "parity": aioserial.PARITY_NONE,  # Other values possible via config
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }

    def __init__(
        self,
        aio: aioserial.AioSerial,
        name="",
    ) -> None:
        """
        Initialize the CVC3000 device controller.

        Parameters:
        -----------
        aio : aioserial.AioSerial
            The serial interface for communication with the CVC3000.
        name : str
            The name assigned to the device instance.
        """
        super().__init__(name)
        self._serial = aio
        self._device_sn: int = None  # type: ignore

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Vacuubrand",
            model="CVC3000",
        )

    @classmethod
    def from_config(cls, port, name=None, **serial_kwargs):
        """
        Create an instance from configuration parameters.

        Used by server to initialize obj from config.

        Only required parameter is 'port'. Optional 'loop' + others (see AioSerial())

        Parameters:
        -----------
        port : str
            The port to which the CVC3000 is connected.
        name : str, optional
            The name assigned to the device instance (default is None).
        **serial_kwargs : dict
            Additional keyword arguments for configuring the serial interface.

        Returns:
        --------
        CVC3000
            An instance of the CVC3000 class.

        Raises:
        -------
        InvalidConfigurationError
            If the serial connection cannot be established.
        """
        # Merge default settings, including serial, with provided ones.
        configuration = CVC3000.DEFAULT_CONFIG | serial_kwargs

        try:
            serial_object = aioserial.AioSerial(port, **configuration)
        except (OSError, aioserial.SerialException) as serial_exception:
            raise InvalidConfigurationError(
                f"Cannot connect to the CVC3000 on the port <{port}>"
            ) from serial_exception

        return cls(serial_object, name)

    async def initialize(self):
        """
        Initialize the connection with the device and configure it.

        This includes setting the device mode, saving configuration, enabling remote control,
        and configuring output settings.

        Raises:
        -------
        InvalidConfigurationError
            If no version reply is received from the CVC3000.
        """
        self.device_info.version = await self.version()
        if not self.device_info.version:
            raise InvalidConfigurationError("No reply received from CVC3000!")

        # Set to CVC3000 mode and save
        await self._send_command_and_read_reply("CVC 3")
        await self._send_command_and_read_reply("STORE")
        # Get reply to set commands
        await self._send_command_and_read_reply("ECHO 1")
        # Remote control
        await self._send_command_and_read_reply("REMOTE 1")
        # mbar, no autostart, no beep, venting auto
        await self._send_command_and_read_reply("OUT_CFG 00001")
        await self.motor_speed(100)

        logger.debug(f"Connected with CVC3000 version {self.device_info.version}")

        self.components.append(CVC3000PressureControl("pressure-control", self))

    async def _send_command_and_read_reply(self, command: str) -> str:
        """
        Send a command to the device and read the reply.

        Parameters:
        -----------
        command : str
            The command string to be transmitted.

        Returns:
        --------
        str
            The reply received from the device.

        Notes:
        ------
        If no reply is received within the timeout period, an error is logged.
        """
        await self._serial.write_async(command.encode("ascii") + b"\r\n")
        logger.debug(f"Command `{command}` sent!")

        # Receive reply and return it after decoding
        try:
            reply = await asyncio.wait_for(self._serial.readline_async(), 2)
        except asyncio.TimeoutError:
            logger.error("No reply received! Unsupported command?")
            return ""

        await asyncio.sleep(0.1)  # Max rate 10 commands/s as per manual

        logger.debug(f"Reply received: {reply}")
        return reply.decode("ascii")

    async def version(self):
        """
        Retrieve the version of the CVC3000 device.

        Returns:
        --------
        str
            The version of the device, or None if the version could not be retrieved.
        """
        raw_version = await self._send_command_and_read_reply("IN_VER")
        # raw_version = CVC 3000 VX.YY
        try:
            return raw_version.split()[-1]
        except IndexError:
            return None

    async def set_pressure(self, pressure: pint.Quantity):
        """
        Set the pressure on the device.

        Parameters:
        -----------
        pressure : pint.Quantity
            The target pressure to be set, expressed in units compatible with the device (e.g., mbar).
        """
        mbar = int(pressure.m_as("mbar"))
        await self._send_command_and_read_reply(f"OUT_SP_1 {mbar}")

    async def get_pressure(self):
        """
        Get the current pressure from the device.

        Returns:
        --------
        float
            The current pressure in mbar.
        """
        pressure_text = await self._send_command_and_read_reply("IN_PV_1")
        return float(pressure_text.split()[0])

    async def motor_speed(self, speed):
        """
        Set the motor speed on the device.

        Parameters:
        -----------
        speed : int
            The target motor speed percentage (0-100%).
        """
        return await self._send_command_and_read_reply(f"OUT_SP_2 {speed}")

    async def status(self) -> ProcessStatus:
        """
        Get the current process status from the device.

        Returns:
        --------
        ProcessStatus
            The status of the device process.
        """
        raw_status = await self._send_command_and_read_reply("IN_STAT")
        # Sometimes fails on first call
        if not raw_status:
            raw_status = await self._send_command_and_read_reply("IN_STAT")
        return ProcessStatus.from_reply(raw_status)



"""Various utility functions for the Magritek device."""
import ctypes
from collections.abc import Callable
from pathlib import Path

from loguru import logger

from flowchem.utils.exceptions import InvalidConfigurationError


def get_my_docs_path() -> Path:
    """Get my docs path on Windows.

    Spinsolve control software is only available on Windows, so lack of cross-platform support is unavoidable.
    XSD and XML schema are installed in my documents, whose location, if custom, can be obtained as follows.
    """
    # From https://stackoverflow.com/questions/6227590

    csidl_personal = 5  # My Documents (csidl = constant special item ID list)
    shgfp_type_current = (
        0  # Get current, not default value (shgfp = shell get folder path)
    )
    buf = ctypes.create_unicode_buffer(2000)
    ctypes.windll.shell32.SHGetFolderPathW(  # type: ignore
        None,
        csidl_personal,
        None,
        shgfp_type_current,
        buf,
    )
    return Path(buf.value)


def create_folder_mapper(
    remote_root: str | Path,
    local_root: str | Path,
) -> Callable[[Path | str], Path]:
    """Return a function that converts path relative to remote_root to their corresponding on local_root.

    Used when using spinsolve on a remote PC to share the result data via a remotely mounted network drive.
    """

    def folder_mapper(path_to_be_translated: Path | str):
        """Convert remote path to the corresponding local location, or None + warning if not possible."""
        # Ensures it is a Path object
        if isinstance(path_to_be_translated, str):
            path_to_be_translated = Path(path_to_be_translated)

        nonlocal remote_root, local_root
        remote_root = Path(remote_root)
        local_root = Path(local_root)
        # If relative translate is not error
        # NOTE: Path.is_relative_to() is available from Py 3.9 only. NBD as this is not often used.
        if not path_to_be_translated.is_relative_to(remote_root):
            logger.exception(
                f"Cannot translate remote path {path_to_be_translated} to a local path!",
            )
            raise InvalidConfigurationError(
                f"Cannot translate remote path {path_to_be_translated} to a local path!"
                f"{path_to_be_translated} is not relative to {remote_root}"
            )

        suffix = path_to_be_translated.relative_to(remote_root)
        return local_root / suffix

    return folder_mapper



"""Pressure sensor."""
from flowchem.devices.flowchem_device import FlowchemDevice

from .sensor import Sensor


class PhotoSensor(Sensor):
    """A photo sensor."""

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(name, hw_device)
        self.add_api_route("/acquire-signal", self.acquire_signal, methods=["GET"])
        self.add_api_route("/calibration", self.calibrate_zero, methods=["PUT"])

    async def calibrate_zero(self):
        """re-calibrate the sensors to their factory zero points."""
        ...

    async def acquire_signal(self):
        """Read from sensor, result to be expressed in % (optional)."""
        ...

    async def power_on(self):
        ...

    async def power_off(self):
        ...



# Add all flowchem-device classes to the flowchem.device namespace
# This is needed by config parser and hides the complexity of the folder hierarchy to the library users.
# All * are defined as __all__ in the corresponding submodule to simplify name changes / refactoring.
from .bronkhorst import *
from .dataapex import *
from .hamilton import *
from .harvardapparatus import *
from .huber import *
from .knauer import *
from .magritek import *
from .manson import *
from .mettlertoledo import *
from .phidgets import *
from .vacuubrand import *
from .vapourtec import *
from .vicivalco import *
from .custom import *


"""Auto-discover the device classes present in the device sub-folders and in the installed plugins."""
import inspect
from importlib.metadata import entry_points
from typing import Any

from loguru import logger

import flowchem.devices
from flowchem.devices.flowchem_device import FlowchemDevice


def is_device_class(test_object):
    """Return true if the object is a subclass of FlowchemDevice."""
    if getattr(test_object, "__module__", None) is None:
        return None
    return (
        inspect.isclass(test_object)
        and issubclass(test_object, FlowchemDevice)
        and test_object.__name__ != "FlowchemDevice"
    )


def _autodiscover_devices_in_module(module) -> dict[str, Any]:
    """Given a module, autodiscover the device classes and return them as dict(name, object)."""
    device_classes = inspect.getmembers(module, is_device_class)
    logger.debug(f"Found {len(device_classes)} device type(s) in {module.__name__}")
    # Dict of device class names and their respective classes, i.e. {device_class_name: DeviceClass}.
    return {obj_class[0]: obj_class[1] for obj_class in device_classes}


def autodiscover_first_party() -> dict[str, Any]:
    """Get classes from `flowchem.devices` subpackages."""
    return _autodiscover_devices_in_module(flowchem.devices)


def autodiscover_third_party() -> dict[str, Any]:
    """Get classes from packages with a `flowchem.devices` entrypoint.

    A plugin structure can be used to add devices from an external package via setuptools entry points.
    See https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/#using-package-metadata
    """
    return {
        k: v
        for ep in entry_points(group="flowchem.devices")
        for (k, v) in _autodiscover_devices_in_module(ep.load()).items()
    }


def autodiscover_device_classes() -> dict[str, Any]:
    """Get all the device-controlling classes, either from `flowchem.devices` or third party packages."""
    first = autodiscover_first_party()
    # logger.info(f"Found {len(first)} 1st-party device type! {list(first.keys())}")
    third = autodiscover_third_party()
    # logger.info(f"Found {len(third)} 3rd-party device type! {list(third.keys())}")

    return third | first  # First party devices will overwrite the third party ones.


if __name__ == "__main__":
    logger.debug(
        f"The following device types were found: {list(autodiscover_device_classes().keys())}",
    )



"""Control module for the Vapourtec R4 heater."""
from __future__ import annotations

from collections import namedtuple
from collections.abc import Iterable

import aioserial
import pint
from loguru import logger

from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.components.technical.temperature import TempRange
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.vapourtec.r4_heater_channel_control import R4HeaterChannelControl
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import dario, jakob, wei_hsin

try:
    # noinspection PyUnresolvedReferences
    from flowchem_vapourtec import VapourtecR4Commands

    HAS_VAPOURTEC_COMMANDS = True
except ImportError:
    HAS_VAPOURTEC_COMMANDS = False


class R4Heater(FlowchemDevice):
    """R4 reactor heater control class."""

    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 19200,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }

    ChannelStatus = namedtuple("ChannelStatus", "state, temperature")

    def __init__(
        self,
        name: str = "",
        min_temp: float | list[float] = -100,
        max_temp: float | list[float] = 250,
        **config,
    ) -> None:
        super().__init__(name)
        # Set min and max temp for all 4 channels
        if not isinstance(min_temp, Iterable):
            min_temp = [min_temp] * 4
        if not isinstance(max_temp, Iterable):
            max_temp = [max_temp] * 4
        assert len(min_temp) == len(max_temp) == 4
        self._min_t = min_temp
        self._max_t = max_temp

        if not HAS_VAPOURTEC_COMMANDS:
            msg = (
                "You tried to use a Vapourtec device but the relevant commands are missing!"
                "Unfortunately, we cannot publish those as they were provided under NDA."
                "Contact Vapourtec for further assistance."
            )
            raise InvalidConfigurationError(
                msg,
            )

        self.cmd = VapourtecR4Commands()

        # Merge default settings, including serial, with provided ones.
        configuration = R4Heater.DEFAULT_CONFIG | config
        try:
            self._serial = aioserial.AioSerial(**configuration)
        except aioserial.SerialException as ex:
            msg = f"Cannot connect to the R4Heater on the port <{config.get('port')}>"
            raise InvalidConfigurationError(
                msg,
            ) from ex

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Vapourtec",
            model="R4 reactor module",
        )

    async def initialize(self):
        """Ensure connection."""
        self.device_info.version = await self.version()
        logger.info(f"Connected with R4Heater version {self.device_info.version}")
        temp_limits = {
            ch_num: TempRange(
                min=ureg.Quantity(f"{t[0]} °C"),
                max=ureg.Quantity(f"{t[1]} °C"),
            )
            for ch_num, t in enumerate(zip(self._min_t, self._max_t, strict=True))
        }
        reactor_positions = [
            R4HeaterChannelControl(f"reactor{n+1}", self, n, temp_limits[n])
            for n in range(4)
        ]
        self.components.extend(reactor_positions)

    async def _write(self, command: str):
        """Write a command to the pump."""
        cmd = command + "\r\n"
        await self._serial.write_async(cmd.encode("ascii"))
        logger.debug(f"Sent command: {command!r}")

    async def _read_reply(self) -> str:
        """Read the pump reply from serial communication."""
        reply_string = await self._serial.readline_async()
        logger.debug(f"Reply received: {reply_string.decode('ascii').rstrip()}")
        return reply_string.decode("ascii")

    async def write_and_read_reply(self, command: str) -> str:
        """Send a command to the pump, read the replies and return it, optionally parsed."""
        self._serial.reset_input_buffer()
        await self._write(command)
        logger.debug(f"Command {command} sent to R4!")
        response = await self._read_reply()

        if not response:
            msg = "No response received from heating module!"
            raise InvalidConfigurationError(msg)

        logger.debug(f"Reply received: {response}")
        return response.rstrip()

    async def version(self):
        """Get firmware version."""
        return await self.write_and_read_reply(self.cmd.VERSION)

    async def set_temperature(self, channel, temperature: pint.Quantity):
        """Set temperature to channel."""
        cmd = self.cmd.SET_TEMPERATURE.format(
            channel=channel,
            temperature_in_C=round(temperature.m_as("°C")),
        )
        await self.write_and_read_reply(cmd)
        # Set temperature implies channel on
        await self.power_on(channel)
        # Verify it is not unplugged
        status = await self.get_status(channel)
        if status.state == "U":
            logger.error(
                f"TARGET CHANNEL {channel} UNPLUGGED! (Note: numbering starts at 0)",
            )

    async def get_status(self, channel) -> ChannelStatus:
        """Get status from channel."""
        # This command is a bit fragile for unknown reasons.
        failure = 0
        while True:
            try:
                raw_status = await self.write_and_read_reply(
                    self.cmd.GET_STATUS.format(channel=channel),
                )
                return R4Heater.ChannelStatus(raw_status[:1], raw_status[1:])
            except InvalidConfigurationError as ex:
                failure += 1
                # Allows 3 failures cause the R4 is choosy at times...
                if failure > 3:
                    raise ex
                else:
                    continue

    async def get_temperature(self, channel):
        """Get temperature (in Celsius) from channel."""
        state = await self.get_status(channel)
        return None if state.temperature == "281.2" else state.temperature

    async def power_on(self, channel):
        """Turn on channel."""
        await self.write_and_read_reply(self.cmd.POWER_ON.format(channel=channel))

    async def power_off(self, channel):
        """Turn off channel."""
        await self.write_and_read_reply(self.cmd.POWER_OFF.format(channel=channel))


if __name__ == "__main__":
    import asyncio

    r4_device = R4Heater(port="COM1")

    async def main(heat):
        """Test function."""
        await heat.initialize()
        # Get reactors
        r1, r2, r3, r4 = heat.components()

        await r1.set_temperature("30 °C")
        print(f"Temperature is {await r1.get_temperature()}")

    asyncio.run(main(r4_device))



"""An NMR control component."""
from fastapi import BackgroundTasks

from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.devices.flowchem_device import FlowchemDevice


class NMRControl(FlowchemComponent):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """NMR Control component."""
        super().__init__(name, hw_device)
        self.add_api_route("/acquire-spectrum", self.acquire_spectrum, methods=["PUT"])
        self.add_api_route("/stop", self.stop, methods=["PUT"])

        # Ontology: fourier transformation NMR instrument
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/OBI_0000487",
        )
        self.component_info.type = "NMR Control"

    async def acquire_spectrum(self, background_tasks: BackgroundTasks):
        """Acquire an NMR spectrum."""
        ...

    async def stop(self):
        """Stop acquisition and exit gracefully."""
        ...



"""This module is used to discover the serial address of any Huber chiller connected to the PC."""
import asyncio
from textwrap import dedent

from loguru import logger

from flowchem.devices.huber.chiller import HuberChiller
from flowchem.utils.exceptions import InvalidConfigurationError


# noinspection PyProtectedMember
def chiller_finder(serial_port) -> set[str]:
    """Try to initialize a Huber chiller on every available COM port."""
    logger.debug(f"Looking for Huber chillers on {serial_port}...")
    dev_config: set[str] = set()

    try:
        chill = HuberChiller.from_config(port=serial_port)
    except InvalidConfigurationError:
        return dev_config

    try:
        asyncio.run(chill.initialize())
    except InvalidConfigurationError:
        chill._serial.close()
        return dev_config

    logger.info(f"Chiller #{chill._device_sn} found on <{serial_port}>")
    dev_config.add(
        dedent(
            f"""
                [device.huber-{chill._device_sn}]
                type = "HuberChiller"
                port = "{serial_port}"\n""",
        ),
    )
    logger.info(f"Close the serial port: <{serial_port}>")
    chill._serial.close()
    return dev_config



"""Autodiscover any supported devices connected to the PC."""
from pathlib import Path

import aioserial
import rich_click as click
from loguru import logger
from serial.tools import list_ports

from flowchem.devices.hamilton.ml600_finder import ml600_finder
from flowchem.devices.harvardapparatus.elite11_finder import elite11_finder
from flowchem.devices.huber.huber_finder import chiller_finder
from flowchem.devices.knauer.knauer_finder import knauer_finder
from flowchem.devices.mettlertoledo.icir_finder import icir_finder
from flowchem.devices.vacuubrand.cvc3000_finder import cvc3000_finder
from flowchem.devices.vapourtec.vapourtec_finder import r4_finder

SERIAL_DEVICE_INSPECTORS = (
    ml600_finder,
    elite11_finder,
    chiller_finder,
    r4_finder,
    cvc3000_finder,
)


def inspect_serial_ports() -> set[str]:
    """Search for known devices on local serial ports and generate config stubs."""
    port_available = [comport.device for comport in list_ports.comports()]
    logger.info(
        f"Found the following serial port(s) on the current device: {port_available}",
    )

    dev_found_config: set[str] = set()
    # Loop each serial port
    for serial_port in port_available:
        logger.info(f"Looking for known devices on {serial_port}...")
        # Check if the serial port is available (i.e. not already open)
        try:
            port = aioserial.Serial(serial_port)
            port.close()
        except OSError:
            logger.info(f"Skipping {serial_port} (cannot be opened: already in use?)")
            continue

        # For each port try all functions that can detect serial port devices
        for inspector in SERIAL_DEVICE_INSPECTORS:
            # a list of config is return by the inspector, if len(config) == 0 then it is falsy
            if config := inspector(serial_port):
                dev_found_config.update(config)
                break
        logger.info(f"No known device found on {serial_port}")

    return dev_found_config


def inspect_eth(source_ip: str) -> set[str]:
    """Search for known devices on ethernet and generate config stubs."""
    logger.info("Starting ethernet detection")
    dev_found_config: set[str] = set()

    # Find Knauer
    knauer_devices = knauer_finder(source_ip)
    dev_found_config.update(knauer_devices)

    # Find iCIR
    icir = icir_finder()
    dev_found_config.update(icir)

    return dev_found_config


@click.command()
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    help="Output file",
    show_default=True,
    default="flowchem_config.toml",
)
@click.option(
    "--overwrite",
    "-w",
    is_flag=True,
    help="Overwrite existing configuration file if present",
)
@click.option("--safe", "-s", is_flag=True, help="Run only safe modules.")
@click.option(
    "--assume-yes",
    "--yes",
    "-y",
    is_flag=True,
    help="Assume 'yes' as answer to all prompts and run non-interactively.",
)
@click.option(
    "--source-ip",
    help="Source IP for broadcast packets. (Relevant if multiple eth interface are available)",
    default="",
)
def main(output, overwrite, safe, assume_yes, source_ip):
    """Auto-find devices connected to the current PC."""
    # Validate output location
    if Path(output).exists() and not overwrite:
        logger.error(
            f"Output file `{output}` already existing! Use `--overwrite` to replace it.",
        )
        return

    # Ask confirmation for serial communication
    confirm = False
    if not safe and not assume_yes:
        logger.warning(
            "The autodiscover include modules that involve communication over serial ports.",
        )
        logger.warning("These modules are *not* guaranteed to be safe!")
        logger.warning(
            "Unsupported devices could be placed in an unsafe state as result of the discovery process!",
        )
        confirm = click.confirm("Do you want to include the search for serial devices?")

    # Search serial devices
    serial_config = (
        inspect_serial_ports() if not safe and (assume_yes or confirm) else set()
    )

    # Search ethernet devices
    eth_config = inspect_eth(source_ip)

    # Print results
    if not serial_config and not eth_config:
        logger.error(f"No device found! The output file `{output}` won't be created.")
        return
    logger.info(f"Found {len(serial_config) + len(eth_config)} devices!")

    # Print configuration
    configuration = "".join(serial_config) + "".join(eth_config)
    logger.info(
        f"The following configuration will be written to `{output}:\n{configuration}",
    )

    # Write to file
    with Path(output).open("w", encoding="utf-8") as f:
        f.write(configuration)
    logger.info(f"Configuration written to `{output}`!")


if __name__ == "__main__":
    main()



"""Knauer valve control."""
import warnings
from enum import Enum
from loguru import logger

from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.knauer._common import KnauerEthernetDevice
from flowchem.devices.knauer.knauer_valve_component import (
    Knauer6PortDistributionValve,
    Knauer12PortDistributionValve,
    Knauer16PortDistributionValve,
    KnauerInjectionValve,
)
from flowchem.utils.exceptions import DeviceError
from flowchem.utils.people import dario, jakob, wei_hsin


class KnauerValveHeads(Enum):
    """Four different valve types can be used. 6port2position injection valve, and 6, 12, 16 multi-position valves."""

    SIX_PORT_TWO_POSITION = "LI"
    SIX_PORT_SIX_POSITION = "6"
    TWELVE_PORT_TWELVE_POSITION = "12"
    SIXTEEN_PORT_SIXTEEN_POSITION = "16"


class KnauerValve(KnauerEthernetDevice, FlowchemDevice):
    """Control Knauer multi position valves.

    Valve type can be 6, 12, 16, or it can be 6 ports, two positions, which will be simply 2 (two states)
    in this case, the response for T is LI. Load and inject can be switched by sending L or I.
    Switching will always be performed by following parent valve type switching commands, so specifying which port
    should be connected
    Regarding initial valve state: If it matters, users want to determine initial state, which then also can happen
    explicit after initialisation and therefore is not deemed critical
    EN: https://www.knauer.net/Dokumente/valves/azura/manuals/v6860_azura_v_2.1s_user-manual_en.pdf
    DE: https://www.knauer.net/Dokumente/valves/azura/manuals/v6860_azura_v_2.1s_benutzerhandbuch_de.pdf
    DIP switch for valve selection
    """

    def __init__(self, ip_address=None, mac_address=None, **kwargs) -> None:
        super().__init__(ip_address, mac_address, **kwargs)
        self.eol = b"\r\n"
        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Knauer",
            model="Valve",
        )

    async def initialize(self):
        """Initialize connection."""
        # The connection is established in KnauerEthernetDevice.initialize()
        await super().initialize()

        # Detect valve type
        self.device_info.additional_info["valve-type"] = await self.get_valve_type()

        # Set components
        valve_component: FlowchemComponent
        match self.device_info.additional_info["valve-type"]:
            case KnauerValveHeads.SIX_PORT_TWO_POSITION:
                valve_component = KnauerInjectionValve("injection-valve", self)
            case KnauerValveHeads.SIX_PORT_SIX_POSITION:
                valve_component = Knauer6PortDistributionValve(
                    "distribution-valve", self
                )
            case KnauerValveHeads.TWELVE_PORT_TWELVE_POSITION:
                valve_component = Knauer12PortDistributionValve(
                    "distribution-valve", self
                )
            case KnauerValveHeads.SIXTEEN_PORT_SIXTEEN_POSITION:
                valve_component = Knauer16PortDistributionValve(
                    "distribution-valve", self
                )
            case _:
                raise RuntimeError("Unknown valve type")
        self.components.append(valve_component)

    @staticmethod
    def handle_errors(reply: str):
        """Return True if there are errors, False otherwise. Warns for errors."""
        if not reply.startswith("E"):
            return

        if "E0" in reply:
            DeviceError(
                "The valve refused to switch.\n"
                "Replace the rotor seals of the valve or replace the motor drive unit.",
            )
        elif "E1" in reply:
            DeviceError(
                "Skipped switch: motor current too high!\n"
                "Replace the rotor seals of the valve.",
            )
        elif "E2" in reply:
            DeviceError(
                "Change from one valve position to the next takes too long.\n"
                "Replace the rotor seals of the valve.",
            )
        elif "E3" in reply:
            DeviceError(
                "Switch position of DIP 3 and 4 are not correct.\n"
                "Correct DIP switch 3 and 4.",
            )
        elif "E4" in reply:
            DeviceError(
                "Valve homing position not recognized.\n" "Adjust sensor board.",
            )
        elif "E5" in reply:
            DeviceError(
                "Switch position of DIP 1 and 2 are not correct.\n"
                "Correct DIP switch 1 and 2.",
            )
        elif "E6" in reply:
            DeviceError("Memory error.\n" "Power-cycle valve!")
        else:
            DeviceError("Unspecified error detected!")

    async def _transmit_and_parse_reply(self, message: str | int) -> str:
        """Send command, receive reply and parse it.

        Args:
        ----
            message (str): command to be sent

        Returns:
        -------
            str: reply
        """
        reply = await self._send_and_receive(str(message))
        self.handle_errors(reply)

        if reply == "?":
            # retry once before failing. This happens often on pos commands!
            reply = await self._send_and_receive(str(message))
            if reply == "?":
                warnings.warn(f"Command failed: {message}")
                logger.warning(f"Command failed: {message}")
                return ""

        return reply

    async def get_valve_type(self):
        """Get valve type, if returned value is not supported throws an error.

        Note that this method is called during initialize(), therefore it is in line
        with the general philosophy of the module to 'fail early' upon init and avoiding
        raising exception after that.
        """
        reply = await self._transmit_and_parse_reply("T")
        assert reply.startswith("VALVE ")
        reply = reply[6:]

        try:
            headtype = KnauerValveHeads(reply)
        except ValueError as value_error:
            raise DeviceError(
                "The valve type returned is not recognized.\n"
                "Are you sure the address provided is correct?\n"
                "Only multi-pos 6, 12, 16 and 2-pos 6 port valves are supported!"
            ) from value_error

        logger.info(f"Valve connected, type: {headtype}.")
        return headtype

    async def get_raw_position(self) -> str:
        """Return current valve position, following valve nomenclature."""
        return await self._transmit_and_parse_reply("P")

    async def set_raw_position(self, position: str) -> bool:
        """Set valve position, following valve nomenclature."""
        return await self._transmit_and_parse_reply(position) != ""


if __name__ == "__main__":
    import asyncio

    v = KnauerValve(ip_address="192.168.1.176")

    async def main(valve):
        """Test function."""
        await valve.initialize()
        await valve.set_raw_position("I")
        print(await valve.get_raw_position())
        await valve.set_raw_position("L")
        print(await valve.get_raw_position())

    asyncio.run(main(v))



"""This module is used to discover the serial address of any CVC3000 connected to the PC."""
import asyncio
from textwrap import dedent

from loguru import logger

from flowchem.devices.vacuubrand.cvc3000 import CVC3000
from flowchem.utils.exceptions import InvalidConfigurationError


# noinspection PyProtectedMember
def cvc3000_finder(serial_port) -> set[str]:
    """Try to initialize a CVC3000 on every available COM port."""
    logger.debug(f"Looking for CVC3000 on {serial_port}...")

    try:
        cvc = CVC3000.from_config(port=serial_port)
    except InvalidConfigurationError:
        return set()

    try:
        asyncio.run(cvc.initialize())
    except InvalidConfigurationError:
        cvc._serial.close()
        return set()

    logger.info(f"CVC3000 {cvc.component_info.version} found on <{serial_port}>")
    dev_config = dedent(
        f"""
            [device.cvc-{cvc._device_sn}]
            type = "CVC3000"
            port = "{serial_port}"\n\n"""
    )
    logger.info(f"Close the serial port: <{serial_port}>")
    cvc._serial.close()
    return set(dev_config)



"""Exceptions used in the flowchem module."""


class DeviceError(BaseException):
    """Generic DeviceError."""


class InvalidConfigurationError(DeviceError):
    """The configuration provided is not valid, e.g. no connection w/ device obtained."""



from __future__ import annotations

from typing import TYPE_CHECKING
from loguru import logger

from flowchem.components.analytics.ir import IRControl, IRSpectrum

if TYPE_CHECKING:
    from .icir import IcIR


class IcIRControl(IRControl):
    """
    IcIR Control component for handling IR spectrometer operations.

    Attributes:
        hw_device (IcIR): The hardware device for the IR spectrometer.
    """
    hw_device: IcIR  # for typing's sake

    def __init__(self, name: str, hw_device: IcIR) -> None:  # type:ignore
        """
        Initialize the IcIRControl component.

        Args:
            name (str): The name of the component.
            hw_device (IcIR): The hardware device instance for controlling the IR spectrometer.
        """
        super().__init__(name, hw_device)
        self.add_api_route("/spectrum-count", self.spectrum_count, methods=["GET"])

    async def acquire_spectrum(self, treated: bool = True) -> IRSpectrum:
        """
        Acquire an IR spectrum from the instrument.

        This method retrieves the most recent infrared (IR) spectrum acquired by the device.
        Depending on the `treated` parameter, it can return either a processed spectrum with background subtraction
        or a raw, unprocessed spectrum.

        The acquisition process works as follows:
        - If `treated` is True, the method returns a spectrum where background subtraction has been performed,
          providing a clean signal suitable for analysis. The treated spectrum is retrieved from the device's
          OPC UA node specified by `SPECTRA_TREATED` ("ns=2;s=Local.iCIR.Probe1.SpectraTreated").
        - If `treated` is False, the method returns the raw, unprocessed spectrum directly from the device,
          without any background correction. The raw spectrum is retrieved from the OPC UA node specified by
          `SPECTRA_RAW` ("ns=2;s=Local.iCIR.Probe1.SpectraRaw").

        Args:
            treated (bool): If True, perform background subtraction and return the treated spectrum.
                            If False, return the raw scan without any processing.

        Returns:
            IRSpectrum: The acquired IR spectrum, either treated or raw.
        """
        if treated:
            return await self.hw_device.last_spectrum_treated()
        else:
            return await self.hw_device.last_spectrum_raw()

    async def spectrum_count(self) -> int:
        """
        Get the count of acquired spectra.

        Returns:
            int: The number of spectra acquired. Returns -1 if the count is None.
        """
        count = await self.hw_device.sample_count()
        if count is not None:
            return count
        else:
            logger.warning("The spectrum count return a 'None' value! This reply was replaced to the int -1.")
            return -1

    async def stop(self):
        """
        Stop the ongoing IR experiment.

        Returns:
            bool: True if the experiment was successfully stopped, False otherwise.
        """
        return await self.hw_device.stop_experiment()



from enum import IntEnum

from pydantic import BaseModel


class PumpControlMode(IntEnum):
    VacuuLAN = 0
    PUMP_DOWN = 1
    VAC_CONTROL = 2
    AUTO = 3
    PROGRAM = 4
    GAUGE = 4


class PumpState(IntEnum):
    OFF = 0
    PUMP_DOWN = 1
    VACUUM_REACHED = 2
    AUTO_OFF = 3  # Below set value


class ProcessStatus(BaseModel):
    is_pump_on: bool
    is_inline_valve_open: bool
    is_coolant_valve_open: bool
    is_venting_valve_open: bool
    control: PumpControlMode
    state: PumpState

    @classmethod
    def from_reply(cls, reply):
        reply_dict = {
            "is_pump_on": bool(int(reply[0])),
            "is_inline_valve_open": bool(int(reply[1])),
            "is_coolant_valve_open": bool(int(reply[2])),
            "is_venting_valve_open": bool(int(reply[3])),
            "control": PumpControlMode(int(reply[4])),
            "state": PumpState(int(reply[5])),
        }
        return cls.model_validate(reply_dict)



"""Base object for all hardware-control device classes."""
from abc import ABC
from collections import namedtuple
from typing import TYPE_CHECKING

from flowchem.components.device_info import DeviceInfo

if TYPE_CHECKING:
    from flowchem.components.flowchem_component import FlowchemComponent

RepeatedTaskInfo = namedtuple("RepeatedTaskInfo", ["seconds_every", "task"])


class FlowchemDevice(ABC):
    """Base flowchem device.

    All hardware-control classes must subclass this to signal they are flowchem-device and be enabled for initialization
    during config parsing.
    """

    def __init__(self, name) -> None:
        """All device have a name, which is the key in the config dict thus unique."""
        self.name = name
        self.device_info = DeviceInfo()
        self.components: list["FlowchemComponent"] = []

    async def initialize(self):
        """Use for setting up async connection to the device, populate components and update device_info with them."""
        pass

    def repeated_task(self) -> RepeatedTaskInfo | None:
        """Use for repeated background task, e.g. session keepalive."""
        return None

    def get_device_info(self) -> DeviceInfo:
        return self.device_info



"""Temperature control, either for heating or cooling."""
from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

import pint
from loguru import logger

from flowchem import ureg
from flowchem.components.flowchem_component import FlowchemComponent

if TYPE_CHECKING:
    from flowchem.devices.flowchem_device import FlowchemDevice


class TempRange(NamedTuple):
    min: pint.Quantity = ureg.Quantity("-100 °C")
    max: pint.Quantity = ureg.Quantity("+250 °C")


class TemperatureControl(FlowchemComponent):
    """A generic temperature controller."""

    def __init__(
        self, name: str, hw_device: FlowchemDevice, temp_limits: TempRange
    ) -> None:
        """Create a TemperatureControl object."""
        super().__init__(name, hw_device)

        self.add_api_route("/temperature", self.set_temperature, methods=["PUT"])
        self.add_api_route("/temperature", self.get_temperature, methods=["GET"])

        self.add_api_route("/power-on", self.power_on, methods=["PUT"])
        self.add_api_route("/power-off", self.power_off, methods=["PUT"])

        self.add_api_route("/target-reached", self.is_target_reached, methods=["GET"])
        # self.add_api_route("/limits", self.temperature_limits, methods=["GET"])

        self._limits = temp_limits

    async def set_temperature(self, temp: str) -> pint.Quantity:
        """Set the target temperature to the given string in "magnitude and unit" format."""
        if temp.isnumeric():
            temp = temp + "°C"
            logger.warning("No units provided to set_temperature, assuming Celsius.")
        set_t = ureg.Quantity(temp)

        if set_t < self._limits[0]:
            set_t = self._limits[0]
            logger.warning(
                f"Temperature requested {set_t} is out of range [{self._limits}] for {self.name}!"
                f"Setting to {self._limits[0]} instead.",
            )

        if set_t > self._limits[1]:
            set_t = self._limits[1]
            logger.warning(
                f"Temperature requested {set_t} is out of range [{self._limits}] for {self.name}!"
                f"Setting to {self._limits[1]} instead.",
            )
        return set_t

    async def get_temperature(self) -> float:  # type: ignore
        """Return temperature in Celsius."""
        ...

    async def is_target_reached(self) -> bool:  # type: ignore
        """Return True if the set temperature target has been reached."""
        ...

    async def temperature_limits(self) -> TempRange:
        """Return a dict with `min` and `max` temperature in Celsius."""
        return self._limits

    async def power_on(self):
        """Turn on temperature control."""
        ...

    async def power_off(self):
        """Turn off temperature control."""
        ...



"""Knauer pump control."""
import asyncio
import warnings
from enum import Enum

import pint
from loguru import logger

from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.knauer._common import KnauerEthernetDevice
from flowchem.devices.knauer.azura_compact_pump import AzuraCompactPump
from flowchem.devices.knauer.azura_compact_sensor import AzuraCompactSensor
from flowchem.utils.exceptions import DeviceError
from flowchem.utils.people import dario, jakob, wei_hsin

FLOW = "FLOW"  # 0-50000 µL/min, int only!
HEADTYPE = "HEADTYPE"  # 10, 50 ml. Value refers to the highest flow rate in ml/min
PMIN10 = "PMIN10"  # 0-400 in 0.1 MPa, use to avoid running dry
PMIN50 = "PMIN50"  # 0-150 in 0.1 MPa, use to avoid running dry
PMAX10 = "PMAX10"  # 0-400 in 0.1 MPa, chosen automatically by selecting pump head
PMAX50 = "PMAX50"  # 0-150 in 0.1 MPa, chosen automatically by selecting pump head
IMIN10 = "IMIN10"  # 0-100 minimum motor current
IMIN50 = "IMIN50"  # 0-100 minimum motor current
STARTLEVEL = "STARTLEVEL"  # 0, 1 configures start. 0 -> only start pump when shorted to GND, 1 -> always allow start
ERRIO = "ERRIO"  # 0, 1 write/read error in/output ??? sets errio either 1 or 0, reports errio:ok
STARTMODE = "STARTMODE"  # 0, 1; 0=pause pump after switch on, 1=start immediately with previous set flow rate
ADJ10 = "ADJ10"  # 100-2000
ADJ50 = "ADJ50"  # 100-2000
CORR10 = "CORR10"  # 0-300
CORR50 = "CORR50"  # 0-300
EXTFLOW = "EXTFLOW?"
IMOTOR = "IMOTOR"  # motor current in relative units 0-100
PRESSURE = "PRESSURE?"  # reads the pressure in 0.1 MPa
ERRORS = "ERRORS"  # displays last 5 error codes
EXTCONTR = "EXTCONTR"  # allows flow control via analog input
LOCAL = "LOCAL"  # no parameter, releases pump to manual control
REMOTE = "REMOTE"  # manual param input prevented
PUMP_ON = "ON"  # starts flow
PUMP_OFF = "OFF"  # stops flow


class AzuraPumpHeads(Enum):
    """Two Pump heads are available for the Azura: 50 mL/min and 10 mL/min."""

    FLOWRATE_FIFTY_ML = 50
    FLOWRATE_TEN_ML = 10


# noinspection DuplicatedCode
class AzuraCompact(KnauerEthernetDevice, FlowchemDevice):
    """
    Control module for Knauer Azura Compact pumps.

    This class handles the initialization, control, and communication with the Knauer Azura Compact pumps.
    It includes methods to set and get parameters such as flow rate, pressure, and pump head type.
    The class supports asynchronous operations to ensure non-blocking communication with the device.

    Attributes:
        ip_address (str): The IP address of the pump.
        mac_address (str): The MAC address of the pump.
        max_pressure (str): Maximum pressure limit for the pump.
        min_pressure (str): Minimum pressure limit for the pump.

    Methods:
        initialize(): Initialize connection and configure the pump.
        error_present(reply: str) -> bool: Check if there are errors in the device's reply.
        create_and_send_command(message, setpoint: int | None = None, setpoint_range: tuple | None = None): Create and
        send a command to the pump.
        get_headtype() -> AzuraPumpHeads: Return the pump's head type.
        set_headtype(head_type: AzuraPumpHeads): Set the pump's head type.
        get_flow_rate() -> float: Get the current flow rate in ml/min.
        set_flow_rate(rate: pint.Quantity): Set the flow rate.
        get_minimum_pressure(): Get the minimum pressure setting.
        set_minimum_pressure(value: str = "0 bar"): Set the minimum pressure.
        get_maximum_pressure() -> str: Get the maximum pressure setting.
        set_maximum_pressure(value: str): Set the maximum pressure.
        set_minimum_motor_current(setpoint=None): Set the minimum motor current.
        is_start_in_required(): Check if START IN is required for the pump to operate.
        require_start_in(value: bool = True): Configure START IN requirement.
        is_autostart_enabled(): Check if autostart is enabled.
        enable_autostart(value: bool = True): Enable or disable autostart.
        get_adjusting_factor(): Get the adjusting factor.
        set_adjusting_factor(setpoint: int | None = None): Set the adjusting factor.
        get_correction_factor(): Get the correction factor.
        set_correction_factor(setpoint=None): Set the correction factor.
        read_pressure() -> pint.Quantity: Read the current pressure from the sensor.
        read_extflow() -> float: Read the external flow rate from analog input.
        read_errors() -> list[int]: Read the last 5 error codes.
        read_motor_current(): Read the motor current in percent.
        infuse(): Start the pump.
        stop(): Stop the pump.
        is_running(): Check if the pump is running.
        set_local(state: bool = True): Relinquish remote control.
        remote_control(state: bool = True): Enable or disable remote control.
        is_analog_control_enabled(): Check if analog control is enabled.
        enable_analog_control(value: bool): Enable or disable external flow control via analog input.
    """

    def __init__(
        self,
        ip_address=None,
        mac_address=None,
        max_pressure: str = "",
        min_pressure: str = "",
        **kwargs,
    ):
        super().__init__(ip_address, mac_address, **kwargs)
        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="knauer",
            model="Azura Compact",
        )
        self.eol = b"\n\r"

        # All the following are set upon initialize()
        self.max_allowed_pressure = 0
        self.max_allowed_flow = 0
        self._headtype = None
        self._running: bool = None  # type: ignore
        self._pressure_max = max_pressure
        self._pressure_min = min_pressure

        self.rate = ureg.parse_expression("0 ml/min")

    async def initialize(self):
        """Initialize connection."""
        # Here the magic happens...
        await super().initialize()

        # Here it is checked that the device is a pump and not a valve
        await self.get_headtype()
        # Place pump in remote control
        await self.remote_control()
        # Also ensure rest state is not pumping.
        await self.stop()

        if self._pressure_max:
            await self.set_maximum_pressure(self._pressure_max)
        if self._pressure_min:
            await self.set_minimum_pressure(self._pressure_min)

        # Set Pump and Sensor components.
        self.components.extend(
            [AzuraCompactPump("pump", self), AzuraCompactSensor("pressure", self)]
        )

    @staticmethod
    def error_present(reply: str) -> bool:
        """Return True if there are errors, False otherwise. Warns for errors."""
        # ERRORS: is the expected answer to read_errors()
        if not reply.startswith("ERROR") or reply.startswith("ERRORS:"):
            return False

        if "ERROR:1" in reply:
            warnings.warn("Invalid message sent to device.\n", stacklevel=2)

        elif "ERROR:2" in reply:
            warnings.warn(
                "Setpoint refused by device.\n" "Refer to manual for allowed values.\n",
                stacklevel=2,
            )
        else:
            warnings.warn("Unspecified error detected!")
        return True

    async def _transmit_and_parse_reply(self, message: str) -> str:
        """Send command and receive reply.

        Deals with all communication based stuff and checks that the valve is of expected type.
        :param message:
        :return: reply: str
        """
        reply = await self._send_and_receive(message)
        if self.error_present(reply):
            return ""

        # Setpoint ok
        if ":OK" in reply:
            logger.debug("setpoint successfully set!")
            return "OK"

        # Replies to 'VALUE?' are in the format 'VALUE:'
        if message[:-1] in reply:
            logger.debug(f"setpoint successfully acquired, value={reply}")
            # Last value after colon
            return reply.split(":")[-1]

        # No reply
        if not reply:
            warnings.warn("No reply received")
            return ""

        warnings.warn(f"Unrecognized reply: {reply}")
        return reply

    async def create_and_send_command(
        self,
        message,
        setpoint: int | None = None,
        setpoint_range: tuple | None = None,
    ):
        """Create and sends a message from the command.

        If setpoint is given, then the command is appended with :value
        If not setpoint is given, a "?" is added for getter syntax

        e.g. message = "HEADTYPE"
        no setpoint -> sends "HEADTYPE?"
        w/ setpoint -> sends "HEADTYPE:<setpoint_value>"
        """
        # GETTER
        if setpoint is None:
            return await self._transmit_and_parse_reply(message + "?")

        # SETTER with range
        if setpoint_range:
            if setpoint in range(*setpoint_range):
                return await self._transmit_and_parse_reply(
                    message + ":" + str(setpoint),
                )

            warnings.warn(
                f"The setpoint provided {setpoint} is not valid for the command "
                f"{message}!\n Accepted range is: {setpoint_range}.\n"
                f"Command ignored",
            )
            return ""

        # SETTER w/o range
        return await self._transmit_and_parse_reply(message + ":" + str(setpoint))

    @property
    def _headtype(self):
        """Return internal head type. Use `set_headtype()` to change in pump."""
        return self.__headtype

    @_headtype.setter
    def _headtype(self, htype):
        self.__headtype = htype

        if htype == AzuraPumpHeads.FLOWRATE_TEN_ML:
            self.max_allowed_pressure, self.max_allowed_flow = 400, 10000
        elif htype == AzuraPumpHeads.FLOWRATE_FIFTY_ML:
            self.max_allowed_pressure, self.max_allowed_flow = 150, 50000

    async def get_headtype(self) -> AzuraPumpHeads:
        """Return pump's head type."""
        head_type_id = await self.create_and_send_command(HEADTYPE)
        try:
            headtype = AzuraPumpHeads(int(head_type_id))
            # Sets internal property (changes max flowrate etc.)
            self._headtype = headtype
        except ValueError as value_error:
            raise DeviceError(
                "It seems you're trying instantiate an unknown device/unknown pump type as Knauer Pump.\n"
                "Only Knauer Azura Compact is supported"
            ) from value_error
        logger.debug(f"Head type of pump {self.ip_address} is {headtype}")

        return headtype

    async def set_headtype(self, head_type: AzuraPumpHeads):
        """Set pump's head type."""
        await self.create_and_send_command(HEADTYPE, setpoint=head_type.value)
        # Update internal property (changes max flowrate etc.)
        self._headtype = head_type
        logger.debug(f"Head type set to {head_type}")

    async def get_flow_rate(self) -> float:
        """Get flow rate in ml/min."""
        flow_value = await self.create_and_send_command(FLOW)
        flowrate = ureg.Quantity(f"{flow_value} ul/min")
        logger.debug(f"Current flow rate is {flowrate}")
        return flowrate.m_as("ml/min")

    async def set_flow_rate(self, rate: pint.Quantity):
        """Set flow rate.

        Args:
        ----
            rate (str): value with units
        """
        await self.create_and_send_command(
            FLOW,
            setpoint=round(rate.m_as("ul/min")),
            setpoint_range=(0, self.max_allowed_flow + 1),
        )
        logger.info(f"Flow set to {rate}")

    async def get_minimum_pressure(self):
        """Get minimum pressure. The pump stops if the measured P is lower than this."""
        command = PMIN10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMIN50
        p_min = await self.create_and_send_command(command) * ureg.bar
        return str(p_min)

    async def set_minimum_pressure(self, value: str = "0 bar"):
        """Set minimum pressure. The pump stops if the measured P is lower than this."""
        pressure = ureg.Quantity(value)
        command = PMIN10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMIN50
        await self.create_and_send_command(
            command,
            setpoint=round(pressure.m_as("bar")),
            setpoint_range=(0, self.max_allowed_pressure + 1),
        )
        logger.info(f"Minimum pressure set to {pressure}")

    async def get_maximum_pressure(self) -> str:
        """Get maximum pressure. The pumps stop if the measured P is higher than this."""
        command = PMAX10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMAX50
        p_max = await self.create_and_send_command(command) * ureg.bar
        return str(p_max)

    async def set_maximum_pressure(self, value: str):
        """Set maximum pressure. The pumps stop if the measured P is higher than this."""
        pressure = ureg.Quantity(value)
        command = PMAX10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMAX50
        await self.create_and_send_command(
            command,
            setpoint=round(pressure.m_as("bar")),
            setpoint_range=(0, self.max_allowed_pressure + 1),
        )
        logger.info(f"Maximum pressure set to {pressure}")

    async def set_minimum_motor_current(self, setpoint=None):
        """Set minimum motor current."""
        command = IMIN10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else IMIN50

        reply = await self.create_and_send_command(
            command,
            setpoint=setpoint,
            setpoint_range=(0, 101),
        )
        logger.debug(f"Minimum motor current set to {setpoint}, returns {reply}")

    async def is_start_in_required(self):
        """Check state of START IN. See require_start_in() for details."""
        runlevel = await self.create_and_send_command(STARTLEVEL)
        return not bool(int(runlevel))

    async def require_start_in(self, value: bool = True):
        """Configure START IN. If required, the pump starts only if the STARTIN pin is shortened to GND.

        True = Pump starts the flow at short circuit contact only. (Start In <> Ground). [0]
        False = Pump starts the flow without a short circuit contact. (Start In <> Ground). [1]
        """
        setpoint = int(not value)
        await self.create_and_send_command(STARTLEVEL, setpoint=setpoint)
        logger.debug(f"Start in required set to {value}")

    async def is_autostart_enabled(self):
        """Return the default behaviour of the pump upon power on."""
        reply = await self.create_and_send_command(STARTMODE)
        return bool(int(reply))

    async def enable_autostart(self, value: bool = True):
        """Set the default behaviour of the pump upon power on.

        :param value: False: pause pump after switch on. True: start pumping with previous flow rate at startup
        :return: device message
        """
        await self.create_and_send_command(STARTMODE, setpoint=int(value))
        logger.debug(f"Autostart set to {value}")

    async def get_adjusting_factor(self):
        """Get the adjust parameter. Not clear what it is."""
        command = ADJ10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else ADJ50
        reply = await self.create_and_send_command(command)
        return int(reply)

    async def set_adjusting_factor(self, setpoint: int | None = None):
        """Set the adjust parameter. Not clear what it is."""
        command = ADJ10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else ADJ50
        reply = await self.create_and_send_command(
            command,
            setpoint=setpoint,
            setpoint_range=(0, 2001),
        )
        logger.debug(f"Adjusting factor of set to {setpoint}, returns {reply}")

    async def get_correction_factor(self):
        """Get the correction factor. Not clear what it is."""
        command = CORR10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else CORR50
        return int(await self.create_and_send_command(command))

    async def set_correction_factor(self, setpoint=None):
        """Set the correction factor. Not clear what it is."""
        command = CORR10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else CORR50
        reply = await self.create_and_send_command(
            command,
            setpoint=setpoint,
            setpoint_range=(0, 301),
        )
        logger.debug(f"Correction factor set to {setpoint}, returns {reply}")

    async def read_pressure(self) -> pint.Quantity:
        """Return pressure if the pump has a pressure sensor."""
        return int(await self._transmit_and_parse_reply(PRESSURE)) * 0.1 * ureg.bar

    async def read_extflow(self) -> float:
        """Read the set flowrate from analog in."""
        ext_flow = await self._transmit_and_parse_reply(EXTFLOW)
        logger.debug(f"Extflow reading returns {ext_flow}")
        return float(ext_flow)

    async def read_errors(self) -> list[int]:
        """Return the last 5 errors."""
        last_5_errors = await self.create_and_send_command(ERRORS)
        logger.debug(f"Error reading returns {last_5_errors}")
        return [int(err_code) for err_code in last_5_errors.split(",")]

    async def read_motor_current(self):
        """Return motor current, relative in percent 0-100."""
        current_percent = int(await self.create_and_send_command(IMOTOR))
        logger.debug(f"Motor current reading returns {current_percent} %")
        return current_percent

    async def infuse(self):
        """Start running pump at the given rate."""
        await self._transmit_and_parse_reply(PUMP_ON)
        self._running = True
        logger.info("Pump started!")
        return True

    async def stop(self):
        """Stop flow."""
        await self._transmit_and_parse_reply(PUMP_OFF)
        self._running = False
        logger.info("Pump stopped")

    def is_running(self):
        """Get pump state."""
        return self._running

    async def set_local(self, state: bool = True):
        """Relinquish remote control."""
        await self.create_and_send_command(LOCAL, setpoint=int(state))
        logger.debug(f"Local control set to {state}")

    async def remote_control(self, state: bool = True):
        """Set remote control on or off."""
        if state:
            await self.create_and_send_command(REMOTE, setpoint=1)
        else:
            await self.create_and_send_command(LOCAL, setpoint=1)
        logger.debug(f"Remote control set to {state}")

    async def is_analog_control_enabled(self):
        """Return the status of the external flow control via analog input."""
        reply = await self.create_and_send_command(EXTCONTR)
        return bool(int(reply))

    async def enable_analog_control(self, value: bool):
        """External flow control via analog input.

        False = prevents external flow control. [0]
        True = allows the flow rate control via analog input 0 - 10V (10ml: 1 V = 1 ml/min, 50ml: 1 V = 5 ml/min). [1]
        """
        await self.create_and_send_command(EXTCONTR, setpoint=int(value))
        logger.debug(f"External control set to {value}")


if __name__ == "__main__":
    # This is a bug of asyncio on Windows :|
    import sys

    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    p = AzuraCompact(ip_address="192.168.1.119")

    async def main(pump: AzuraCompact):
        """Test function."""
        await pump.initialize()
        c = pump.components
        print(c)
        pc: AzuraCompactPump = c[0]  # type:ignore
        print(pc)
        print(await pc.infuse(rate="0.1 ml/min"))
        await pump.set_flow_rate(ureg.Quantity("0.1 ml/min"))
        await pump.infuse()
        await asyncio.sleep(5)
        await pump.stop()
        print(await pc.is_pumping())

    asyncio.run(main(p))



"""Vici valve component."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .vici_valve import ViciValve
from flowchem.components.valves.injection_valves import SixPortTwoPositionValve


class ViciInjectionValve(SixPortTwoPositionValve):
    hw_device: ViciValve  # for typing's sake

    # todo this needs to be adapted to new code
    def _change_connections(self, raw_position: int | str, reverse: bool = False) -> str:
        # TODO maybe needs addition of one, not sure
        if not reverse:
            translated = str(raw_position)
        else:
            translated = str(raw_position)
        return translated



from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem import ureg
from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.components.sensors.pressure_sensor import PressureSensor
from flowchem.devices.flowchem_device import FlowchemDevice

if TYPE_CHECKING:
    from .el_flow import EPC, MFC


class EPCComponent(PressureSensor):
    """
    A class to represent a component that interfaces with an Electronic Pressure Controller (EPC) device.

    Attributes:
    -----------
    hw_device : EPC
        The hardware device (EPC) this component interfaces with.

    Methods:
    --------
    set_pressure_setpoint(pressure: str) -> bool:
        Set the controlled pressure to the instrument; default unit is bar.
    get_pressure() -> float:
        Get the current system pressure in bar.
    stop() -> bool:
        Stop the pressure controller by setting pressure to 0 bar.
    read_pressure(units: str = "bar"):
        Read the current pressure from the sensor and return it in the specified units.
    """
    hw_device: EPC  # just for typing

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Constructs all the necessary attributes for the EPCComponent object.

        Parameters:
        -----------
        name : str
            The name of the EPC component.
        hw_device : FlowchemDevice
            The hardware device (EPC) this component interfaces with.
        """
        super().__init__(name, hw_device)
        self.add_api_route("/get-pressure", self.get_pressure, methods=["GET"])
        self.add_api_route("/set-pressure", self.set_pressure_setpoint, methods=["PUT"])
        self.add_api_route("/stop", self.stop, methods=["PUT"])

    async def set_pressure_setpoint(self, pressure: str) -> bool:
        """
        Set the controlled pressure on the instrument. The default unit is bar.
        The unit is handled through the `ureg` package to convert it to the unit configured on the device.
        To standardize units, it is recommended that the user works with the bar unit.

        Parameters:
        -----------
        pressure : str
            The desired pressure to set.

        Returns:
        --------
        bool
            True if the pressure setpoint was set successfully.
        """
        await self.hw_device.set_pressure(pressure)
        return True

    async def get_pressure(self) -> float:
        """
        Get the current system pressure in bar.

        Returns:
        --------
        float
            The current pressure in bar.
        """
        return await self.hw_device.get_pressure()

    async def stop(self) -> bool:
        """
        Stop the pressure controller by setting pressure to 0 bar.

        Returns:
        --------
        bool
            True if the pressure controller was stopped successfully.
        """
        await self.hw_device.set_pressure("0 bar")
        return True

    async def read_pressure(self, units: str = "bar"):
        """
        Read the current pressure from the sensor and return it in the specified units.

        Parameters:
        -----------
        units : str, optional
            The units in which to return the pressure (default is bar).

        Returns:
        --------
        float
            The current pressure in the specified units.
        """
        p = await self.hw_device.get_pressure()
        return p * ureg(units)  # <Quantity(4.56, 'bar')>


class MFCComponent(FlowchemComponent):
    """
    A class to represent a Mass Flow Controller (MFC) component.

    Attributes:
    -----------
    hw_device : MFC
        The hardware device (MFC) this component interfaces with.

    Methods:
    --------
    set_flow_setpoint(flowrate: str) -> bool:
        Set the flow rate to the instrument; default unit is ml/min.
    get_flow_setpoint() -> float:
        Get the current flow rate in ml/min.
    stop() -> bool:
        Stop the mass flow controller.
    """
    hw_device: MFC  # just for typing

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Constructs all the necessary attributes for the MFCComponent object.

        Parameters:
        -----------
        name : str
            The name of the MFC component.
        hw_device : FlowchemDevice
            The hardware device (MFC) this component interfaces with.
        """
        super().__init__(name, hw_device)
        self.add_api_route("/get-flow-rate", self.get_flow_setpoint, methods=["GET"])
        self.add_api_route("/set-flow-rate", self.set_flow_setpoint, methods=["PUT"])
        self.add_api_route("/stop", self.stop, methods=["PUT"])

    async def set_flow_setpoint(self, flowrate: str) -> bool:
        """
        Set the flow rate to the instrument; default unit is ml/min.

        Parameters:
        -----------
        flowrate : str
            The desired flow rate to set.

        Returns:
        --------
        bool
            True if the flow rate setpoint was set successfully.
        """
        await self.hw_device.set_flow_setpoint(flowrate)
        return True

    async def get_flow_setpoint(self) -> float:
        """
        Get the current flow rate in ml/min.

        Returns:
        --------
        float
            The current flow rate in ml/min.
        """
        return await self.hw_device.get_flow_setpoint()

    async def stop(self) -> bool:
        """
        Stop the mass flow controller.

        Returns:
        --------
        bool
            True if the mass flow controller was stopped successfully.
        """
        await self.hw_device.set_flow_setpoint("0 ml/min")
        return True



# Single-sourcing version, Option 5 in https://packaging.python.org/en/latest/guides/single-sourcing-package-version/
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

import pint

# Unit registry
ureg = pint.UnitRegistry(autoconvert_offset_to_baseunit=True)
ureg.define("step = []")
ureg.define("stroke = 48000 * step")



"""Phidget-based devices."""
from .bubble_sensor import PhidgetBubbleSensor, PhidgetPowerSource5V
from .pressure_sensor import PhidgetPressureSensor

__all__ = ["PhidgetPressureSensor", "PhidgetBubbleSensor", "PhidgetPowerSource5V"]



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

import inspect

from fastapi import APIRouter
from loguru import logger

from flowchem.components.component_info import ComponentInfo

if TYPE_CHECKING:
    from flowchem.devices.flowchem_device import FlowchemDevice


class FlowchemComponent:
    """
    A base class for Flowchem components that integrates with the decomposition of the hardware device.

    This class provides the foundational setup for creating abstract components of hardware devices. It forms the
    foundation to expose API routes for interacting with the hardware through a hierarchical structure.

    Attributes:
    -----------
    name : str
        The name of the component.
    hw_device : FlowchemDevice
        The hardware device instance associated with this component.
    component_info : ComponentInfo
        Metadata about the component.
    _router : APIRouter
        The API router for the component to define HTTP endpoints.

    Methods:
    --------
    add_api_route(path: str, endpoint: Callable, **kwargs):
        Add an API route to the component's router.
    get_component_info() -> ComponentInfo:
        Retrieve the component's metadata.
    """
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Initialize the FlowchemComponent with a name and associated hardware device.

        Parameters:
        -----------
        name : str
            The name assigned to this component.
        hw_device : FlowchemDevice
            The hardware device instance associated with this component.
        """
        self.name = name
        self.hw_device = hw_device
        self.component_info = ComponentInfo(
            name=name,
            parent_device=self.hw_device.name,
            corresponding_class=[cls.__name__ for cls in inspect.getmro(self.__class__)]
        )

        # Initialize router
        self._router = APIRouter(
            prefix=f"/{self.component_info.parent_device}/{name}",
            tags=[self.component_info.parent_device],
        )
        self.add_api_route(
            "/",
            self.get_component_info,
            methods=["GET"],
            response_model=ComponentInfo,
        )

    @property
    def router(self):
        """
        Return the API router.

        This serves as a hook for subclasses to add their specific routes.

        Returns:
        --------
        APIRouter
            The API router instance.
        """
        return self._router

    def add_api_route(self, path: str, endpoint: Callable, **kwargs):
        """
        Add an API route to the component's router.

        This method allows subclasses to define their own API endpoints.

        Parameters:
        -----------
        path : str
            The URL path for the API route.
        endpoint : Callable
            The function to be called when the route is accessed.
        kwargs : dict
            Additional arguments to configure the route.
        """
        logger.debug(f"Adding route {path} for router of {self.name}")
        self._router.add_api_route(path, endpoint, **kwargs)

    def get_component_info(self) -> ComponentInfo:
        """
        Retrieve the component's metadata.

        This endpoint provides information about the component, such as its name and associated hardware device.

        Returns:
        --------
        ComponentInfo
            Metadata about the component.
        """
        return self.component_info




"""Knauer valve component."""
from __future__ import annotations

from typing import TYPE_CHECKING
from enum import Enum

if TYPE_CHECKING:
    from .knauer_valve import KnauerValve
from flowchem.components.valves.distribution_valves import (
    SixPortDistributionValve,
    SixteenPortDistributionValve,
    TwelvePortDistributionValve,
)
from flowchem.components.valves.injection_valves import SixPortTwoPositionValve


class KnauerInjectionValve(SixPortTwoPositionValve):
    """
    Control a Knauer Injection Valve.

    Attributes:
        hw_device (KnauerValve): The hardware device for the Knauer valve.
    """
    hw_device: KnauerValve  # for typing's sake

    def __init__(self, name: str, hw_device: KnauerValve) -> None:
        """
        Initialize the KnauerInjectionValve component.

        Args:
            name (str): The name of the component.
            hw_device (KnauerValve): The hardware device instance for controlling the Knauer valve.
        """
        super().__init__(name, hw_device)

        self.add_api_route("/monitor_position", self.get_monitor_position, methods=["GET"])
        self.add_api_route("/monitor_position", self.set_monitor_position, methods=["PUT"])

    class LoadInject(Enum):
        LOAD = "L"
        INJECT = "I"

    def _change_connections(self, raw_position: str | int, reverse: bool = False):
        position_mapping = {0: "L", 1: "I"}
        if reverse:
            return str([key for key, value in position_mapping.items() if value == raw_position][0])
        else:
            if type(raw_position) is int:
                return position_mapping[raw_position]
            else:
                raise TypeError

    async def get_monitor_position(self) -> str:
        """Get current valve position."""
        pos = self.LoadInject(await self.hw_device.get_raw_position())
        return pos.name

    async def set_monitor_position(self, position: str):
        """Move valve to position."""
        try:
            return await self.hw_device.set_raw_position(self.LoadInject[position.upper()].value)
        except KeyError as e:
            raise Exception(f"Please give allowed positions {[pos.name for pos in self.LoadInject]}") from e


class Knauer6PortDistributionValve(SixPortDistributionValve):
    """KnauerValve of type SIX_PORT_SIX_POSITION."""

    hw_device: KnauerValve  # for typing's sake

    def __init__(self, name: str, hw_device: KnauerValve) -> None:
        """Create a ValveControl object."""
        super().__init__(name, hw_device)

        self.add_api_route("/monitor_position", self.get_monitor_position, methods=["GET"])
        self.add_api_route("/monitor_position", self.set_monitor_position, methods=["PUT"])

    def _change_connections(self, raw_position: int | str, reverse: bool = False):
        """
        Change connections based on the valve's raw position.
        The raw_position is the device-specific position naming

        Args:
            raw_position (int or str): The raw position of the valve.
            reverse (bool): Whether to reverse the mapping.
            For this class the raw_position is the same of the monitor position

        Returns:
            int: The mapped position.
        """
        if reverse:
            return int(raw_position) - 1
        else:
            return int(raw_position) + 1

    async def get_monitor_position(self) -> str:
        return await self.hw_device.get_raw_position()

    async def set_monitor_position(self, position: str):
        return await self.hw_device.set_raw_position(position)


class Knauer12PortDistributionValve(TwelvePortDistributionValve):
    """KnauerValve of type SIX_PORT_SIX_POSITION."""

    hw_device: KnauerValve  # for typing's sake

    def __init__(self, name: str, hw_device: KnauerValve) -> None:
        """Create a ValveControl object."""
        super().__init__(name, hw_device)

        self.add_api_route("/monitor_position", self.get_monitor_position, methods=["GET"])
        self.add_api_route("/monitor_position", self.set_monitor_position, methods=["PUT"])

    def _change_connections(self, raw_position: int | str, reverse: bool = False):
        """
        Change connections based on the valve's raw position.

        Args:
            raw_position (int): The raw position of the valve.
            reverse (bool): Whether to reverse the mapping.

        Returns:
            int: The mapped position.
        """
        if reverse:
            return int(raw_position) - 1
        else:
            return int(raw_position) + 1

    async def get_monitor_position(self) -> str:
        return await self.hw_device.get_raw_position()

    async def set_monitor_position(self, position: str):
        return await self.hw_device.set_raw_position(position)


class Knauer16PortDistributionValve(SixteenPortDistributionValve):
    """KnauerValve of type SIX_PORT_SIX_POSITION."""

    hw_device: KnauerValve  # for typing's sake

    def __init__(self, name: str, hw_device: KnauerValve) -> None:
        """Create a ValveControl object."""
        super().__init__(name, hw_device)

        self.add_api_route("/monitor_position", self.get_monitor_position, methods=["GET"])
        self.add_api_route("/monitor_position", self.set_monitor_position, methods=["PUT"])

    def _change_connections(self, raw_position: int | str, reverse: bool = False):
        if reverse:
            return int(raw_position) - 1
        else:
            return int(raw_position) + 1

    async def get_monitor_position(self) -> str:
        return await self.hw_device.get_raw_position()

    async def set_monitor_position(self, position: str):
        return await self.hw_device.set_raw_position(position)



from pydantic import NameEmail

__all__ = ["dario", "jakob", "wei_hsin", "samuel_saraiva"]


dario = NameEmail(name="Dario Cambiè", email="2422614+dcambie@users.noreply.github.com")
jakob = NameEmail(name="Jakob Wolf", email="Jakob.Wolf@mpikg.mpg.de")
wei_hsin = NameEmail(name="Wei-Hsin Hsu", email="Wei-hsin.Hsu@mpikg.mpg.de")
miguel = NameEmail(name="Miguel Garcia", email="miguel.garcia@mpikg.mpg.de")
samuel_saraiva = NameEmail(name="Samuel Saraiva", email="Samuel.Saraiva@mpikg.mpg.de")




"""Functions related to instrument response parsing."""
import warnings
from enum import Enum

from lxml import etree


class StatusNotification(Enum):
    """Represent the type of the status notification."""

    STARTED = 1  # <State status="Ready"> received, starting protocol
    RUNNING = 2  # All good, <Progress> received, protocol is running
    STOPPING = 3  # Abort called, waiting for current scan end
    FINISHING = 4  # Upon <State status="Ready": acquisition over, processing ongoing
    COMPLETED = 5  # Upon <Completed>: with this also processing/saving data is over
    ERROR = 6  # If an error occurs
    UNKNOWN = 7


def parse_status_notification(xml_message: etree._Element):
    """Parse a status notification reply."""
    status_notification = xml_message.find(".//StatusNotification")
    assert status_notification is not None, "a StatusNotification tree is needed"

    # StatusNotification child can be <State> (w/ submsg), <Progress>, <Completed> or <Error>
    match status_notification[0].tag, status_notification[0].get("status"):
        case ["State", "Running"]:
            status = StatusNotification.STARTED
        case ["State", "Ready"]:
            status = StatusNotification.FINISHING
        case ["State", "Stopping"]:
            status = StatusNotification.STOPPING
        case ["Progress", None]:
            status = StatusNotification.RUNNING
        case ["Completed", None]:
            status = StatusNotification.COMPLETED
        case ["Error", None]:
            status = StatusNotification.ERROR
        case _:
            warnings.warn("Could not recognize StatusNotification state!")
            status = StatusNotification.UNKNOWN

    return status, status_notification[0].get("dataFolder")



from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import BackgroundTasks

from flowchem.components.analytics.nmr import NMRControl

if TYPE_CHECKING:
    from .spinsolve import Spinsolve


class SpinsolveControl(NMRControl):
    hw_device: Spinsolve  # for typing's sake

    def __init__(self, name: str, hw_device: Spinsolve) -> None:  # type:ignore
        """HPLC Control component. Sends methods, starts run, do stuff."""
        super().__init__(name, hw_device)
        # Solvent
        self.add_api_route("/solvent", self.hw_device.get_solvent, methods=["GET"])
        self.add_api_route("/solvent", self.hw_device.set_solvent, methods=["PUT"])
        # Sample name
        self.add_api_route("/sample-name", self.hw_device.get_sample, methods=["GET"])
        self.add_api_route("/sample-name", self.hw_device.set_sample, methods=["PUT"])
        # User data
        self.add_api_route("/user-data", self.hw_device.get_user_data, methods=["GET"])
        self.add_api_route("/user-data", self.hw_device.set_user_data, methods=["PUT"])
        # Protocols
        self.add_api_route(
            "/protocol-list",
            self.hw_device.list_protocols,
            methods=["GET"],
        )
        self.add_api_route(
            "/spectrum-folder",
            self.hw_device.get_result_folder,
            methods=["GET"],
        )
        self.add_api_route(
            "/is-busy",
            self.hw_device.is_protocol_running,
            methods=["GET"],
        )

    async def acquire_spectrum(
        self,
        background_tasks: BackgroundTasks,
        protocol="H",
        options=None,
    ) -> int:
        """Acquire an NMR spectrum.

        Return an ID to be passed to get_result_folder, it will return the result folder after acquisition end.
        """
        return await self.hw_device.run_protocol(
            name=protocol,
            background_tasks=background_tasks,
            options=options,
        )

    async def stop(self):
        return await self.hw_device.abort()



"""Knauer devices."""
from .azura_compact import AzuraCompact
from .dad import KnauerDAD
from .knauer_finder import knauer_finder
from .knauer_valve import KnauerValve

__all__ = [
    "knauer_finder",
    "AzuraCompact",
    "KnauerDAD",
    "KnauerValve"
]



"""Use Phidgets to control lab devices. So far, only 0-5 volt interface for bubble-sensor.

PhidgetBubbleSensor_power control the power of the bubble sensor
PhidgetBubbleSensor measure the signal of the bubble sensor

"""
import time

from loguru import logger

from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.phidgets.bubble_sensor_component import (
    PhidgetBubbleSensorComponent,
    PhidgetBubbleSensorPowerComponent,
)
from flowchem.utils.people import dario, jakob, wei_hsin

try:
    from Phidget22.Devices.DigitalOutput import DigitalOutput  # power source
    from Phidget22.Devices.VoltageInput import PowerSupply, VoltageInput  # Sensor
    from Phidget22.PhidgetException import PhidgetException

    HAS_PHIDGET = True
except ImportError:
    HAS_PHIDGET = False

from flowchem.utils.exceptions import (
    InvalidConfigurationError,
)  # configuration is not valid


class PhidgetPowerSource5V(FlowchemDevice):
    """Use a Phidget power source to apply power to the sensor."""

    def __init__(
        self,
        vint_serial_number: int = -1,
        vint_hub_port: int = -1,
        vint_channel: int = -1,
        phidget_is_remote: bool = False,
        name: str = "",
    ) -> None:
        """Initialize BubbleSensor with the given voltage range (sensor-specific!)."""
        super().__init__(name=name)
        if not HAS_PHIDGET:
            raise InvalidConfigurationError(
                "Phidget unusable: library or package not installed."
            )

        # power switch
        self.phidget = DigitalOutput()

        # Ensure connection with the right sensor (ideally these are from config)
        if vint_serial_number > -1:
            self.phidget.setDeviceSerialNumber(vint_serial_number)
        if vint_hub_port > -1:
            self.phidget.setHubPort(vint_hub_port)
            self.phidget.setIsHubPortDevice(True)
        if vint_channel > -1:
            self.phidget.setChannel(vint_channel)

        # Fancy remote sensors?
        if phidget_is_remote:
            from Phidget22.Net import Net
            from Phidget22.PhidgetServerType import PhidgetServerType

            Net.enableServerDiscovery(PhidgetServerType.PHIDGETSERVER_DEVICEREMOTE)
            self.phidget.setIsRemote(True)

        try:
            self.phidget.openWaitForAttachment(1000)
            logger.debug("power of tube sensor is connected!")
        except PhidgetException as pe:
            raise InvalidConfigurationError(
                "Cannot connect to sensor! Check it is not already opened elsewhere and settings..."
            ) from pe

        # Set power supply to 5V to provide power
        self.phidget.setDutyCycle(0.0)
        logger.info("power of tube sensor is turn off!")
        # self.phidget.setState(True)  #setting DutyCycle to 1.0

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Phidget",
            model="VINT",
            serial_number=vint_serial_number,
        )

    async def initialize(self):
        self.components.append(PhidgetBubbleSensorPowerComponent("5V", self))

    def __del__(self) -> None:
        """Ensure connection closure upon deletion."""
        self.phidget.close()

    def power_on(self):
        """Control the power of the device."""
        self.phidget.setDutyCycle(1.0)  # self.phidget.setState(True)
        logger.debug("tube sensor power is turn on!")

    def power_off(self):
        self.phidget.setState(False)
        logger.debug("tube sensor power is turn off!")

    def is_attached(self) -> bool:
        """Whether the device is connected."""
        return bool(self.phidget.getAttached())

    def is_poweron(self) -> bool:
        """Whether the power is on."""
        return bool(self.phidget.getState())


class PhidgetBubbleSensor(FlowchemDevice):
    """Use a Phidget voltage input to translate a Tube Liquid Sensor OPB350 5 Voltage signal
    to the corresponding light penetration value.
    """

    def __init__(
        self,
        vint_serial_number: int = -1,
        vint_hub_port: int = -1,
        vint_channel: int = -1,
        phidget_is_remote: bool = False,
        data_interval: int = 250,  # ms
        name: str = "",
    ) -> None:
        """Initialize BubbleSensor with the given voltage range (sensor-specific!)."""
        super().__init__(name=name)
        if not HAS_PHIDGET:
            raise InvalidConfigurationError(
                "Phidget unusable: library or package not installed."
            )

        # Sensor range
        # sensor_min, sensor_max = intensity_range
        # self._min_intensity = sensor_min
        # self._max_intensity = sensor_max

        # Voltage meter by Versatile input Phidget DAQ1400_0
        self.phidget = VoltageInput()

        # Ensure connection with the right sensor (ideally these are from config)
        if vint_serial_number > -1:
            self.phidget.setDeviceSerialNumber(vint_serial_number)
        if vint_hub_port > -1:
            self.phidget.setHubPort(vint_hub_port)
        if vint_channel > -1:
            self.phidget.setChannel(vint_channel)

        # Fancy remote sensors?
        if phidget_is_remote:
            from Phidget22.Net import Net
            from Phidget22.PhidgetServerType import PhidgetServerType

            Net.enableServerDiscovery(PhidgetServerType.PHIDGETSERVER_DEVICEREMOTE)
            self.phidget.setIsRemote(True)

        try:
            self.phidget.openWaitForAttachment(1000)
            logger.debug("tube sensor is connected!")
        except PhidgetException as pe:
            raise InvalidConfigurationError(
                "Cannot connect to sensor! Check it is not already opened elsewhere and settings..."
            ) from pe

        # Set power supply to 12V to start measurement
        self.phidget.setPowerSupply(PowerSupply.POWER_SUPPLY_OFF)
        logger.info(
            f"default tube sensor is turn off, default data interval is {data_interval} ms!"
        )
        self.phidget.setDataInterval(data_interval)

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Phidget",
            model="VINT",
            serial_number=vint_serial_number,
        )

    async def initialize(self):
        self.components.append(PhidgetBubbleSensorComponent("bubble-sensor", self))

    def __del__(self) -> None:
        """Ensure connection closure upon deletion."""
        self.phidget.close()

    def power_on(self):
        """Turn on the measurement of the bubble sensor."""
        self.phidget.setPowerSupply(PowerSupply.POWER_SUPPLY_12V)
        logger.debug("measurement of tube sensor is turn on!")

    def power_off(self):
        """Turn off the supply to stop measurement."""
        self.phidget.setPowerSupply(PowerSupply.POWER_SUPPLY_OFF)
        logger.debug("measurement of tube sensor is turn off!")

    def is_attached(self) -> bool:
        """Whether the device is connected."""
        return bool(self.phidget.getAttached())

    def get_dataInterval(self) -> int:
        """Get Data Interval form the initial setting."""
        return self.phidget.getDataInterval()

    def set_dataInterval(self, datainterval: int) -> None:
        """Set Data Interval: 20-6000 ms."""
        self.phidget.setDataInterval(datainterval)
        logger.debug(f"change data interval to {datainterval}!")

    def _voltage_to_intensity(self, voltage_in_volt: float) -> float:
        """Convert current reading into percentage."""
        intensity_reading = voltage_in_volt * 20
        logger.debug(f"Read intensity {intensity_reading}!")
        return intensity_reading

    def read_voltage(self) -> float:  # type: ignore
        """Read intensity from the sensor and returns it as float."""
        try:
            voltage = self.phidget.getVoltage()
            logger.debug(f"Actual voltage: {voltage}")
            return voltage
        except PhidgetException:
            logger.error("Cannot read intensity!")
            return 0

    def read_intensity(self) -> float:  # type: ignore
        """Read intensity from voltage."""
        voltage = self.read_voltage()
        return self._voltage_to_intensity(voltage)

    # def getMaxVoltage(self):
    # https: // www.phidgets.com /?view = api


if __name__ == "__main__":
    # turn on the  power of the bubble tube
    power = PhidgetPowerSource5V(
        vint_serial_number=627768,
        vint_hub_port=3,
        vint_channel=0,
    )

    # turn on the sensor
    BubbleSensor_1 = PhidgetBubbleSensor(
        vint_serial_number=627768,
        vint_hub_port=0,
        vint_channel=0,
    )

    while True:
        print(BubbleSensor_1.read_intensity())
        time.sleep(0.1)



"""Huber TemperatureControl component."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .chiller import HuberChiller


from flowchem.components.technical.temperature import TemperatureControl


class HuberTemperatureControl(TemperatureControl):
    """
    Temperature control component for Huber Chillers.

    Attributes:
        hw_device (HuberChiller): The hardware device controlling the temperature.
    """
    hw_device: HuberChiller  # for typing's sake

    async def set_temperature(self, temp: str):
        """
        Set the target temperature to the given value.

        Args:
            temp (str): The desired temperature as a string in "magnitude and unit" format (scalar + unit), e.g. "7 °C"

        Returns:
            bool: True if the temperature was successfully set, False otherwise.
        """
        set_t = await super().set_temperature(temp)
        return await self.hw_device.set_temperature(set_t)

    async def get_temperature(self) -> float:
        """
        Get the current temperature.

        Returns:
            float: The current temperature in Celsius.
        """
        return await self.hw_device.get_temperature()

    async def is_target_reached(self) -> bool:
        """
        Check if the set temperature target has been reached.

        Returns:
            bool: True if the target temperature has been reached, False otherwise.
        """
        return await self.hw_device.target_reached()

    async def power_on(self):
        """
        Turn on the temperature control.

        Returns:
            bool: True if the command was successfully sent, False otherwise.
        """
        return await self.hw_device._send_command_and_read_reply("{M140001")

    async def power_off(self):
        """
        Turn off the temperature control.

        Returns:
            bool: True if the command was successfully sent, False otherwise.
        """
        return await self.hw_device._send_command_and_read_reply("{M140000")



from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import Query

from flowchem.components.analytics.hplc import HPLCControl

if TYPE_CHECKING:
    from flowchem.devices import Clarity


class ClarityComponent(HPLCControl):
    """
    A component for controlling a ClarityChrom system with HPLC functionality.

    This class extends HPLCControl and adds ClarityChrom-specific functionality, such as starting and stopping the ClarityChrom instance and sending methods or running samples.

    Attributes:
    -----------
    hw_device : Clarity
        The ClarityChrom hardware device this component interfaces with.

    Methods:
    --------
    exit() -> bool:
        Exit the ClarityChrom application.
    send_method(method_name: str) -> bool:
        Set the HPLC method using a file with a .MET extension.
    run_sample(sample_name: str, method_name: str) -> bool:
        Run an analysis on the instrument with the specified sample and method.
    """

    hw_device: Clarity  # for typing's sake

    def __init__(self, name: str, hw_device: Clarity) -> None:
        """
        Constructs all the necessary attributes for the ClarityComponent object.

        Parameters:
        -----------
        name : str
            The name of the Clarity component.
        hw_device : Clarity
            The ClarityChrom hardware device this component interfaces with.
        """
        super().__init__(name, hw_device)
        # Clarity-specific command
        self.add_api_route("/exit", self.exit, methods=["PUT"])

    async def exit(self) -> bool:
        """
        Exit the ClarityChrom application.

        Returns:
        --------
        bool
            True if the exit command was successfully executed, False otherwise.
        """
        return await self.hw_device.execute_command("exit", without_instrument_num=True)

    async def send_method(
        self,
        method_name: str = Query(
            default=...,
            description="Name of the method file",
            examples=["MyMethod.MET"],
            alias="method-name",
        ),
    ) -> bool:
        """
        Set the HPLC method using a file with a .MET extension.

        Ensure that the 'Send Method to Instrument' option is selected in the Method Sending Options dialog in
        System Configuration.

        Parameters:
        -----------
        method_name : str
            The name of the method file to be sent.

        Returns:
        --------
        bool
            True if the method was successfully sent, False otherwise.
        """
        return await self.hw_device.execute_command(f" {method_name}")

    async def run_sample(
        self,
        sample_name: str = Query(
            default=...,
            description="Sample name",
            examples=["JB-123-crude-2h"],
            alias="sample-name",
        ),
        method_name: str = Query(
            default=...,
            description="Name of the method file",
            examples=["MyMethod.MET"],
            alias="method-name",
        ),
    ) -> bool:
        """
        Run an analysis on the instrument with the specified sample and method.

        Note that it takes at least 2 sec until the run actually starts (depending on instrument configuration).
        While the export of the chromatogram in e.g. ASCII format can be achieved programmatically via the CLI, the best
        solution is to enable automatic data export for all runs of the HPLC as the chromatogram will be automatically
        exported as soon as the run is finished.

        Parameters:
        -----------
        sample_name : str
            The name of the sample to be run.
        method_name : str
            The name of the method file to be used.

        Returns:
        --------
        bool
            True if the sample was successfully run, False otherwise.
        """
        if not await self.hw_device.execute_command(f'set_sample_name="{sample_name}"'):
            return False
        if not await self.send_method(method_name):
            return False
        return await self.hw_device.execute_command(
            f"run={self.hw_device.instrument}",
            without_instrument_num=True,
        )



"""Injection valves are multiport, two-position valves, e.g. 6-2 commonly used w/ injection loops for HPLC injection."""
from flowchem.components.valves.valve import Valve
from flowchem.devices.flowchem_device import FlowchemDevice


class SixPortTwoPositionValve(Valve):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        # These are hardware-port, only input and output are routable from the fixed syringe.
        # All three are listed as this simplifies the creation of graphs
        super().__init__(
            name,
            hw_device,
            stator_ports=[(1, 2, 3, 4, 5, 6), ()],
            rotor_ports=[(7, 7, 8, 8, 9, 9), ()],
        )



"""Distribution valves, generally connected to syringe pumps, direct the flow from a fixed port to one of the others."""
from flowchem.components.valves.valve import Valve
from flowchem.devices.flowchem_device import FlowchemDevice


class TwoPortDistributionValve(Valve):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(name, hw_device,
                         stator_ports=[(1, 2), (0,)],
                         rotor_ports=[(3, None), (3,)],
                         )


class FourPortDistributionValve(Valve):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(name, hw_device,
                         stator_ports=[(1, 2, 3, 4), (0,)],
                         rotor_ports=[(5, None, None, None), (5,)],
                         )


class SixPortDistributionValve(Valve):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(
            name,
            hw_device,
            stator_ports=[(1, 2, 3, 4, 5, 6), (0,)],
            rotor_ports=[(7, None, None, None, None, None), (7,)],
        )


class TwelvePortDistributionValve(Valve):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(
            name,
            hw_device,
            stator_ports=[(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12), (0,)],
            rotor_ports=[(13, None, None, None, None, None, None, None, None, None, None, None), (13,)],
        )


class SixteenPortDistributionValve(Valve):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(
            name,
            hw_device,
            stator_ports=[(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16), (0,)],
            rotor_ports=[(17, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None),
                         (17,)],
        )


# tot this should be 4 port sth valve
class ThreePortFourPositionValve(Valve):
    """
    This is of type HamiltonDualPumpValveOnRight
    """

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(
            name,
            hw_device,
            stator_ports=[(None, 1, 2, 3,), (0,)],
            rotor_ports=[(4, 4, 5, 5), (4,)],
        )

# tot this should be 4 port
class ThreePortTwoPositionValve(Valve):
    """
    This is of type HamiltonDualPumpValveOnLeft
    """

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(
            name,
            hw_device,
            stator_ports=[(None, 1, 2, 3,), (0,)],
            rotor_ports=[(4, 4, None, None), (None,)],
        )

class FourPortFivePositionValve(Valve):
    """
    This is of type HamiltonDualPumpValveOnLeft
    """
    # rotor and stator look confusing, however this is necessary to apply the logic
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(
            name,
            hw_device,
            stator_ports=[(None, None, 1, None, 2, None, 3, None,), (0,)],
            rotor_ports=[(None, 5, None, None, 4, None, 4, None), (5,)],
        )




import asyncio
from typing import Any

from loguru import logger
from zeroconf import Zeroconf
from zeroconf.asyncio import AsyncServiceBrowser, AsyncServiceInfo

from flowchem.client.client import FlowchemCommonDeviceListener
from flowchem.client.common import (
    FLOWCHEM_TYPE,
    device_url_from_service_info,
    flowchem_devices_from_url_dict,
    zeroconf_name_to_device_name,
)
from flowchem.client.device_client import FlowchemDeviceClient


class FlowchemAsyncDeviceListener(FlowchemCommonDeviceListener):
    bg_tasks: set[Any] = set()

    async def _resolve_service(self, zc: Zeroconf, type_: str, name: str):
        service_info = AsyncServiceInfo(type_, name)
        await service_info.async_request(zc, 3000)
        if service_info:
            device_name = zeroconf_name_to_device_name(name)
            if url := device_url_from_service_info(service_info, device_name):
                self.flowchem_devices[device_name] = url
        else:
            logger.warning(f"No info for service {name}!")

    def _save_device_info(self, zc: Zeroconf, type_: str, name: str, active_ips=None) -> None:
        task = asyncio.ensure_future(self._resolve_service(zc, type_, name))
        self.bg_tasks.add(task)
        task.add_done_callback(self.bg_tasks.discard)


async def async_get_all_flowchem_devices(
    timeout: float = 3000,
) -> dict[str, FlowchemDeviceClient]:
    """Search for flowchem devices and returns them in a dict (key=name, value=IPv4Address)."""
    listener = FlowchemAsyncDeviceListener()
    browser = AsyncServiceBrowser(Zeroconf(), FLOWCHEM_TYPE, listener)
    await asyncio.sleep(timeout / 1000)
    await browser.async_cancel()

    return flowchem_devices_from_url_dict(listener.flowchem_devices)


if __name__ == "__main__":

    async def main():
        dev_info = await async_get_all_flowchem_devices()
        print(dev_info)

    asyncio.run(main())
